"""Test the _extract_the_func functionality in agents_wrapper."""

import pytest

pytest.importorskip("agents")  # Skip if agents not installed

from agents import function_tool

from veris_ai.agents_wrapper import _extract_the_func, _find_user_function_in_closure


def test_extract_simple_function():
    """Test extracting a simple user function from FunctionTool."""

    def my_function(x: int, y: int) -> int:
        """Simple test function."""
        return x + y

    tool = function_tool(my_function)
    extracted = _extract_the_func(tool.on_invoke_tool)

    assert extracted is not None
    assert extracted is my_function
    assert extracted.__name__ == "my_function"
    assert extracted(2, 3) == 5


def test_extract_async_function():
    """Test extracting an async function from FunctionTool."""

    async def my_async_function(text: str) -> str:
        """Async test function."""
        return f"processed: {text}"

    tool = function_tool(my_async_function)
    extracted = _extract_the_func(tool.on_invoke_tool)

    assert extracted is not None
    assert extracted is my_async_function
    assert extracted.__name__ == "my_async_function"


def test_extract_function_with_context():
    """Test extracting a function that takes a context parameter."""
    from agents import RunContextWrapper

    def my_context_function(ctx: RunContextWrapper, value: int) -> int:
        """Function with context."""
        return value * 2

    tool = function_tool(my_context_function)
    extracted = _extract_the_func(tool.on_invoke_tool)

    assert extracted is not None
    assert extracted is my_context_function
    assert extracted.__name__ == "my_context_function"


def test_extract_lambda_function():
    """Test extracting a lambda function (edge case)."""
    # Lambda functions might behave differently
    lambda_func = lambda x, y: x + y  # noqa: E731
    tool = function_tool(lambda_func)
    extracted = _extract_the_func(tool.on_invoke_tool)

    assert extracted is not None
    assert extracted is lambda_func
    # Lambda functions have a generic name
    assert extracted(3, 4) == 7


@pytest.mark.skip(reason="Agents library has issues with decorated functions and strict schema")
def test_extract_decorated_function():
    """Test extracting a decorated function."""

    def my_decorator(func):
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    @my_decorator
    def decorated_function(x: int) -> int:
        """Decorated function."""
        return x * 2

    tool = function_tool(decorated_function)
    extracted = _extract_the_func(tool.on_invoke_tool)

    assert extracted is not None
    # Might be the wrapper or the original depending on decorator
    assert callable(extracted)
    assert extracted(5) == 10


def test_find_user_function_filters_library_code():
    """Test that _find_user_function_in_closure filters out library code."""
    import json

    # Create a closure with both user and library functions
    def user_func():
        return "user"

    library_func = json.loads  # A library function

    closure = []
    for func in [user_func, library_func]:
        # Create a closure cell-like structure
        # In reality, this is done by the Python interpreter
        # but for testing we simulate it
        class Cell:
            def __init__(self, contents):
                self.cell_contents = contents

        closure.append(Cell(func))

    # Convert to tuple to match closure structure
    closure = tuple(closure)

    result = _find_user_function_in_closure(closure)

    # Should find user_func, not json.loads
    assert result is user_func


def test_extract_returns_none_for_invalid_input():
    """Test that _extract_the_func returns None for invalid input."""

    def regular_function():
        pass

    # Not a FunctionTool's on_invoke_tool
    result = _extract_the_func(regular_function)
    assert result is None

    # None input
    result = _extract_the_func(None)
    assert result is None


def test_extract_handles_complex_closures():
    """Test extraction with nested closures."""

    def outer():
        outer_var = 10

        def middle():
            middle_var = 20

            def inner(x: int) -> int:
                return x + outer_var + middle_var

            return inner

        return middle()

    complex_func = outer()
    tool = function_tool(complex_func)
    extracted = _extract_the_func(tool.on_invoke_tool)

    assert extracted is not None
    assert extracted is complex_func
    assert extracted(5) == 35  # 5 + 10 + 20



"""Helper functions for tests."""

import base64
import json


def create_test_token(
    session_id: str = "test-session", thread_id: str | None = "test-thread"
) -> str:
    """Create a base64-encoded test token.

    Args:
        session_id: Session ID to include in token
        thread_id: Thread ID to include in token (optional)

    Returns:
        Base64-encoded JSON token
    """
    token_data = {"session_id": session_id}
    if thread_id is not None:
        token_data["thread_id"] = thread_id

    return base64.b64encode(json.dumps(token_data).encode("utf-8")).decode("utf-8")



"""Tests for token decoding functionality."""

import base64
import json

import pytest

from veris_ai import veris


class TestTokenDecoding:
    """Test token decoding and session/thread ID extraction."""

    def test_decode_valid_token_with_both_ids(self):
        """Test decoding a valid token with both session_id and thread_id."""
        session_id = "test-session-123"
        thread_id = "thread-456"
        token_data = {"session_id": session_id, "thread_id": thread_id}
        token = base64.b64encode(json.dumps(token_data).encode("utf-8")).decode("utf-8")

        veris.set_session_id(token)

        assert veris.session_id == session_id
        assert veris.thread_id == thread_id

        veris.clear_session_id()

    def test_decode_valid_token_with_only_session_id(self):
        """Test decoding a token with only session_id (no thread_id)."""
        session_id = "test-session-789"
        token_data = {"session_id": session_id}
        token = base64.b64encode(json.dumps(token_data).encode("utf-8")).decode("utf-8")

        veris.set_session_id(token)

        assert veris.session_id == session_id
        assert veris.thread_id is None

        veris.clear_session_id()

    def test_decode_token_with_uuid_format(self):
        """Test decoding token with UUID-formatted IDs."""
        session_id = "550e8400-e29b-41d4-a716-446655440000"
        thread_id = "6ba7b810-9dad-11d1-80b4-00c04fd430c8"
        token_data = {"session_id": session_id, "thread_id": thread_id}
        token = base64.b64encode(json.dumps(token_data).encode("utf-8")).decode("utf-8")

        veris.set_session_id(token)

        assert veris.session_id == session_id
        assert veris.thread_id == thread_id

        veris.clear_session_id()

    def test_decode_invalid_base64_raises_error(self):
        """Test that invalid base64 raises ValueError."""
        invalid_token = "not-valid-base64!!!"

        with pytest.raises(ValueError, match="Invalid token format"):
            veris.set_session_id(invalid_token)

    def test_decode_invalid_json_raises_error(self):
        """Test that valid base64 but invalid JSON raises ValueError."""
        invalid_json = base64.b64encode(b"not valid json").decode("utf-8")

        with pytest.raises(ValueError, match="Invalid token format"):
            veris.set_session_id(invalid_json)

    def test_decode_non_object_json_raises_error(self):
        """Test that JSON array instead of object raises ValueError."""
        json_array = base64.b64encode(
            json.dumps(["array", "not", "object"]).encode("utf-8")
        ).decode("utf-8")

        with pytest.raises(ValueError, match="Token must decode to a JSON object"):
            veris.set_session_id(json_array)

    def test_decode_missing_session_id_raises_error(self):
        """Test that token without session_id raises ValueError."""
        token_data = {"thread_id": "thread-only"}
        token = base64.b64encode(json.dumps(token_data).encode("utf-8")).decode("utf-8")

        with pytest.raises(ValueError, match="Token must contain 'session_id' field"):
            veris.set_session_id(token)

    def test_clear_session_id_clears_both_values(self):
        """Test that clear_session_id clears both session_id and thread_id."""
        session_id = "test-session"
        thread_id = "test-thread"
        token_data = {"session_id": session_id, "thread_id": thread_id}
        token = base64.b64encode(json.dumps(token_data).encode("utf-8")).decode("utf-8")

        veris.set_session_id(token)
        assert veris.session_id == session_id
        assert veris.thread_id == thread_id

        veris.clear_session_id()
        assert veris.session_id is None
        assert veris.thread_id is None

    def test_decode_token_with_additional_fields(self):
        """Test that additional fields in token are ignored."""
        session_id = "test-session"
        thread_id = "test-thread"
        token_data = {
            "session_id": session_id,
            "thread_id": thread_id,
            "extra_field": "ignored",
            "another": 123,
        }
        token = base64.b64encode(json.dumps(token_data).encode("utf-8")).decode("utf-8")

        veris.set_session_id(token)

        assert veris.session_id == session_id
        assert veris.thread_id == thread_id

        veris.clear_session_id()

    def test_decode_token_with_empty_strings(self):
        """Test decoding token with empty string values."""
        session_id = ""
        thread_id = ""
        token_data = {"session_id": session_id, "thread_id": thread_id}
        token = base64.b64encode(json.dumps(token_data).encode("utf-8")).decode("utf-8")

        veris.set_session_id(token)

        assert veris.session_id == session_id
        assert veris.thread_id == thread_id

        veris.clear_session_id()



"""Example of using the Veris SDK with OpenAI agents to intercept tool calls."""

import asyncio
import os
import sys

from veris_ai import veris, Runner, VerisConfig

# Ensure the openai-agents package is installed
try:
    from agents import Agent, function_tool
except ImportError:
    print("Please install the openai-agents package: pip install veris-ai[agents]")
    sys.exit(1)

# Note: The Veris Runner extends OpenAI's Runner, providing a seamless upgrade path


# Define some example tools
@function_tool
def calculator(x: int, y: int, operation: str = "add") -> int:
    """Performs arithmetic operations on two numbers."""
    if operation == "add":
        return x + y
    if operation == "multiply":
        return x * y
    if operation == "subtract":
        return x - y
    if operation == "divide":
        return x // y if y != 0 else 0
    return 0


@function_tool
def search_web(query: str) -> str:
    """Searches the web for information."""
    return f"Search results for: {query}"


@function_tool
def get_weather(location: str) -> str:
    """Gets current weather for a location."""
    return f"Weather in {location}: Sunny, 72°F"


async def example_basic_wrap() -> None:
    """Basic example: Intercept all tools using Runner (recommended)."""
    print("\n" + "=" * 60)
    print("Example 1: Basic - All Tools Intercepted (New Runner Class)")
    print("=" * 60)

    # Create an agent with tools
    agent = Agent(
        name="Assistant",
        model="gpt-4",
        tools=[calculator, search_web, get_weather],
        instructions="You are a helpful assistant with various tools.",
    )

    # Use the new Runner class (recommended approach)
    result = await Runner.run(agent, "What's 10 + 5? Also search for Python tutorials.")
    print(f"Result: {result.output}")


async def example_selective_wrap() -> None:
    """Example with selective tool interception using Runner with config."""
    print("\n" + "=" * 60)
    print("Example 2: Selective - Only Specific Tools (With Config)")
    print("=" * 60)

    agent = Agent(
        name="Assistant",
        model="gpt-4",
        tools=[calculator, search_web, get_weather],
        instructions="You are a helpful assistant.",
    )

    print("📝 Only 'calculator' and 'search_web' will be mocked via Veris")
    print("   'get_weather' will run normally")

    # Create configuration and use Runner.run
    config = VerisConfig(include_tools=["calculator", "search_web"])
    result = await Runner.run(agent, "Calculate 5+3 and check weather in NYC", veris_config=config)
    print(f"Result: {result.output}")


async def example_exclude_tools() -> None:
    """Example excluding specific tools from interception."""
    print("\n" + "=" * 60)
    print("Example 3: Exclude Tools - Keep Some Tools Normal")
    print("=" * 60)

    agent = Agent(
        name="Assistant",
        model="gpt-4",
        tools=[calculator, search_web, get_weather],
        instructions="You are a helpful assistant.",
    )

    print("📝 All tools EXCEPT 'get_weather' will be mocked via Veris")

    # Use Runner.run with exclude_tools config
    config = VerisConfig(exclude_tools=["get_weather"])
    result = await Runner.run(
        agent, "Calculate 10*5 and check weather in London", veris_config=config
    )
    print(f"Result: {result.output}")


async def example_simple_calculation() -> None:
    """Example with simple calculation tool."""
    print("\n" + "=" * 60)
    print("Example 4: Simple Calculation (Direct Runner.run)")
    print("=" * 60)

    agent = Agent(
        name="Assistant",
        model="gpt-4",
        tools=[calculator],
        instructions="You are a helpful assistant.",
    )

    # Use Runner.run directly (like OpenAI's Runner)
    result = await Runner.run(agent, "What's 7 * 8?")
    print(f"Result: {result.output}")


async def main() -> None:
    """Run all examples."""
    print("\n🚀 VERIS SDK - OPENAI AGENTS WRAPPER EXAMPLES")
    print("\nThese examples demonstrate different ways to intercept tool calls")
    print("from OpenAI agents and route them through the Veris SDK.\n")
    print("🆕 Using the Veris Runner class that extends OpenAI's Runner!\n")

    # Set up the environment
    # In production, these would be set as environment variables
    if not os.getenv("VERIS_ENDPOINT_URL"):
        print("⚠️  VERIS_ENDPOINT_URL not set. Using demo endpoint.")
        os.environ["VERIS_ENDPOINT_URL"] = "http://demo.veris.ai"

    # Set a session ID to enable mocking
    veris.set_session_id("example-session-123")
    print(f"✅ Session ID set: {veris.session_id}")

    # Check if OpenAI API key is set
    if not os.getenv("OPENAI_API_KEY"):
        print("\n⚠️  Note: OPENAI_API_KEY not set.")
        print("   The examples will show the pattern but won't actually run.")
        print("   Set your API key to see real agent execution.\n")

    try:
        # Run examples
        await example_basic_wrap()
        await example_selective_wrap()
        await example_exclude_tools()
        await example_simple_calculation()
    except Exception as e:
        print(f"\n❌ Error: {e}")
        print("\nMake sure to:")
        print("1. Set OPENAI_API_KEY environment variable")
        print("2. Set VERIS_ENDPOINT_URL to your Veris endpoint")
        print("3. Install dependencies: pip install veris-ai[agents]")
    finally:
        # Clear session when done
        veris.clear_session_id()
        print("\n✅ Session cleared")

    print("\n✨ Examples complete!")
    print("\nNext steps:")
    print("1. Set up your Veris endpoint")
    print("2. Configure your OpenAI API key")
    print("3. Use the Runner class from veris_ai to intercept tool calls")
    print("\n💡 Usage:")
    print("   from veris_ai import Runner, VerisConfig")
    print("   # Direct usage without config")
    print("   result = await Runner.run(agent, 'your input')")
    print("   # With configuration")
    print("   config = VerisConfig(include_tools=['tool1'])")
    print("   result = await Runner.run(agent, 'your input', veris_config=config)")


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import inspect
import logging
import sys
import types
import typing
from contextlib import suppress
from typing import (
    Any,
    ForwardRef,
    Literal,
    NotRequired,
    Required,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)
from collections.abc import Callable

from pydantic import BaseModel

logger = logging.getLogger(__name__)


def convert_to_type(value: object, target_type: type) -> object:
    """Convert a value to the specified type."""
    # Special case: Any type returns value as-is
    if target_type is Any:
        return value

    origin = get_origin(target_type)

    # Define conversion strategies for different type origins
    converters = {
        list: _convert_list,
        dict: _convert_dict,
        Union: _convert_union,
        types.UnionType: _convert_union,  # Handle Python 3.10+ union syntax (str | int)
    }

    # Use appropriate converter based on origin
    if origin in converters:
        return converters[origin](value, target_type)

    # Handle primitives and custom types
    return _convert_simple_type(value, target_type)


def _convert_list(value: object, target_type: type) -> list:
    """Convert a value to a typed list."""
    if not isinstance(value, list):
        error_msg = f"Expected list but got {type(value)}"
        raise ValueError(error_msg)

    item_type = get_args(target_type)[0]
    return [convert_to_type(item, item_type) for item in value]


def _convert_dict(value: object, target_type: type) -> dict:
    """Convert a value to a typed dict."""
    if not isinstance(value, dict):
        error_msg = f"Expected dict but got {type(value)}"
        raise ValueError(error_msg)

    key_type, value_type = get_args(target_type)
    return {convert_to_type(k, key_type): convert_to_type(v, value_type) for k, v in value.items()}


def _convert_union(value: object, target_type: type) -> object:
    """Try to convert value to one of the union types."""
    union_types = get_args(target_type)

    for possible_type in union_types:
        with suppress(ValueError, TypeError):
            return convert_to_type(value, possible_type)

    error_msg = f"Could not convert {value} to any of the union types {union_types}"
    raise ValueError(error_msg)


def _convert_simple_type(value: object, target_type: type) -> object:
    """Convert to primitive or custom types."""
    # Primitive types
    if target_type in (str, int, float, bool):
        return target_type(value)

    # Custom types - try kwargs for dicts, then direct instantiation
    if isinstance(value, dict):
        with suppress(TypeError):
            return target_type(**value)

    if target_type is types.NoneType:
        return None

    return target_type(value)


def _resolve_forward_ref(ref: ForwardRef, module_context: types.ModuleType | None = None) -> Any:  # noqa: ANN401
    """Resolve a ForwardRef to its actual type."""
    if not isinstance(ref, ForwardRef):
        return ref

    # Try to evaluate the forward reference
    try:
        # Get the module's namespace for evaluation
        namespace = dict(vars(module_context)) if module_context else {}

        # Add common typing imports to namespace
        namespace.update(
            {
                "Union": Union,
                "Any": Any,
                "Literal": Literal,
                "Required": Required,
                "NotRequired": NotRequired,
                "List": list,
                "Dict": dict,
                "Optional": typing.Optional,
                "Iterable": typing.Iterable,
                "str": str,
                "int": int,
                "float": float,
                "bool": bool,
            },
        )

        # Try to import from the same module to resolve local references
        if module_context and hasattr(module_context, "__name__"):
            with suppress(Exception):
                # Import all from the module to get access to local types
                exec(f"from {module_context.__name__} import *", namespace)  # noqa: S102

        # Get the forward reference string
        ref_string = ref.__forward_arg__ if hasattr(ref, "__forward_arg__") else str(ref)

        # Try to evaluate the forward reference string
        return eval(ref_string, namespace, namespace)  # noqa: S307
    except Exception:
        # If we can't resolve it, return the ref itself
        return ref


def _unwrap_required(field_type: Any) -> tuple[Any, bool]:  # noqa: ANN401
    """Unwrap Required/NotRequired and return the inner type and whether it's required."""
    origin = get_origin(field_type)

    # Check if it's Required or NotRequired
    if origin is Required:
        args = get_args(field_type)
        return args[0] if args else field_type, True
    if origin is NotRequired:
        args = get_args(field_type)
        return args[0] if args else field_type, False

    # Default to required for TypedDict fields
    return field_type, True


def extract_json_schema(target_type: Any) -> dict:  # noqa: PLR0911, PLR0912, C901, ANN401
    """Extract the JSON schema from a type or pydantic model.

    Args:
        target_type: The type or pydantic model to extract the JSON schema from.

    Returns:
        A dictionary representing the JSON schema.

    Example:
        >>> extract_json_schema(int)
        {"type": "integer"}

        >>> extract_json_schema(list[int])
        {"type": "array", "items": {"type": "integer"}}

        >>> extract_json_schema(list[User])
        {"type": "array", "items": {"type": "object", "properties": {...}}}
    """
    # Handle Pydantic BaseModel instances or classes
    if isinstance(target_type, type) and issubclass(target_type, BaseModel):
        return target_type.model_json_schema()
    if isinstance(target_type, BaseModel):
        return target_type.model_json_schema()

    # Handle TypedDict
    if (
        isinstance(target_type, type)
        and hasattr(target_type, "__annotations__")
        and hasattr(target_type, "__total__")
    ):
        # This is a TypedDict
        properties = {}
        required = []

        # Get the module context for resolving forward references
        module = sys.modules.get(target_type.__module__)

        for field_name, field_type_annotation in target_type.__annotations__.items():
            # Resolve forward references if present
            resolved_type = field_type_annotation
            if isinstance(resolved_type, ForwardRef):
                resolved_type = _resolve_forward_ref(resolved_type, module)

            # Unwrap Required/NotRequired
            unwrapped_type, is_required = _unwrap_required(resolved_type)

            # Extract schema for the unwrapped type
            properties[field_name] = extract_json_schema(unwrapped_type)

            # Add to required list if necessary
            if is_required and getattr(target_type, "__total__", True):
                required.append(field_name)

        schema = {"type": "object", "properties": properties}
        if required:
            schema["required"] = required
        return schema

    # Handle built-in types
    type_mapping = {
        str: {"type": "string"},
        int: {"type": "integer"},
        float: {"type": "number"},
        bool: {"type": "boolean"},
        type(None): {"type": "null"},
        Any: {},  # Empty schema for Any type
    }

    if target_type in type_mapping:
        return type_mapping[target_type]

    # Handle generic types
    origin = get_origin(target_type)

    # Handle bare collection types
    if target_type is list:
        return {"type": "array"}
    if target_type is dict:
        return {"type": "object"}
    if target_type is tuple:
        return {"type": "array"}

    # Handle Literal types
    if origin is Literal:
        values = get_args(target_type)
        if len(values) == 1:
            # Single literal value - use const
            return {"const": values[0]}
        # Multiple literal values - use enum
        return {"enum": list(values)}

    if origin is list:
        args = get_args(target_type)
        if args:
            return {"type": "array", "items": extract_json_schema(args[0])}
        return {"type": "array"}

    if origin is dict:
        args = get_args(target_type)
        if len(args) == 2:  # noqa: PLR2004
            # For typed dicts like dict[str, int]
            return {
                "type": "object",
                "additionalProperties": extract_json_schema(args[1]),
            }
        return {"type": "object"}

    if origin is Union:
        args = get_args(target_type)
        # Handle Optional types (Union[T, None])
        if len(args) == 2 and type(None) in args:  # noqa: PLR2004
            non_none_type = args[0] if args[1] is type(None) else args[1]
            schema = extract_json_schema(non_none_type)
            return {"anyOf": [schema, {"type": "null"}]}
        # Handle general Union types
        return {"anyOf": [extract_json_schema(arg) for arg in args]}

    if origin is tuple:
        args = get_args(target_type)
        if args:
            return {
                "type": "array",
                "prefixItems": [extract_json_schema(arg) for arg in args],
                "minItems": len(args),
                "maxItems": len(args),
            }
        return {"type": "array"}

    # Default case for unknown types
    return {"type": "object"}


def get_function_parameters(
    func: Callable, args: tuple[object, ...], kwargs: dict[str, object]
) -> dict[str, dict[str, str]]:
    """Get the parameters for a function."""
    params_info = {}
    sig = inspect.signature(func)
    bound_args = sig.bind(*args, **kwargs)
    bound_args.apply_defaults()
    _ = bound_args.arguments.pop("ctx", None)
    _ = bound_args.arguments.pop("self", None)
    _ = bound_args.arguments.pop("cls", None)
    for param_name, param_value in bound_args.arguments.items():
        params_info[param_name] = {
            "value": str(param_value),
            "type": str(get_type_hints(func).get(param_name, Any)),
        }
    return params_info


# Callback utility functions


def get_input_parameters(func: Callable, args: tuple, kwargs: dict) -> dict[str, Any]:
    """Get the actual input parameters for callbacks.

    Args:
        func: The function being called
        args: Positional arguments
        kwargs: Keyword arguments

    Returns:
        Dictionary of parameter names to their actual values (not stringified),
        excluding self and cls parameters. Preserves ctx if present.
    """
    sig = inspect.signature(func)
    bound_args = sig.bind(*args, **kwargs)
    bound_args.apply_defaults()

    # Remove only self and cls - preserve ctx and all other parameters
    params = dict(bound_args.arguments)
    params.pop("self", None)
    params.pop("cls", None)

    return params


def filter_callback_parameters(callback: Callable, params: dict[str, Any]) -> dict[str, Any]:
    """Filter parameters to match what the callback can accept.

    Args:
        callback: The callback function to inspect
        params: Dictionary of all available parameters

    Returns:
        Filtered dictionary containing only parameters the callback accepts
    """
    try:
        sig = inspect.signature(callback)

        # Check if callback accepts **kwargs (VAR_KEYWORD parameter)
        has_var_keyword = any(
            p.kind == inspect.Parameter.VAR_KEYWORD for p in sig.parameters.values()
        )

        # If callback accepts **kwargs, pass all parameters
        if has_var_keyword:
            return params

        # Otherwise, filter to only include parameters the callback accepts
        accepted_params = {}
        for param_name in sig.parameters:
            if param_name in params:
                accepted_params[param_name] = params[param_name]

        return accepted_params
    except (ValueError, TypeError):
        # If we can't inspect the signature, pass all parameters
        # and let the callback handle it (will fail if incompatible)
        return params


def execute_callback(
    callback: Callable | None,
    data: Any,  # noqa: ANN401
    unpack: bool = False,
) -> None:
    """Execute a callback synchronously if provided.

    Args:
        callback: The callback callable to execute
        data: The data to pass to the callback
        unpack: If True and data is a dict, unpack it as keyword arguments

    Note:
        Exceptions in callbacks are caught and logged to prevent breaking the main flow.
    """
    if callback is None:
        return

    try:
        if unpack and isinstance(data, dict):
            # Filter parameters to match callback signature
            filtered_data = filter_callback_parameters(callback, data)
            callback(**filtered_data)
        else:
            callback(data)
    except Exception as e:
        logger.warning(f"Callback execution failed: {e}", exc_info=True)


async def execute_callback_async(
    callback: Callable | None,
    data: Any,  # noqa: ANN401
    unpack: bool = False,
) -> None:
    """Execute a callback asynchronously if provided.

    Handles both sync and async callback callables.

    Args:
        callback: The callback callable to execute (can be sync or async)
        data: The data to pass to the callback
        unpack: If True and data is a dict, unpack it as keyword arguments

    Note:
        Exceptions in callbacks are caught and logged to prevent breaking the main flow.
    """
    if callback is None:
        return

    try:
        if inspect.iscoroutinefunction(callback):
            if unpack and isinstance(data, dict):
                # Filter parameters to match callback signature
                filtered_data = filter_callback_parameters(callback, data)
                await callback(**filtered_data)
            else:
                await callback(data)
        else:
            if unpack and isinstance(data, dict):
                # Filter parameters to match callback signature
                filtered_data = filter_callback_parameters(callback, data)
                result = callback(**filtered_data)
            else:
                result = callback(data)
            # If the result is a coroutine (can happen with functools.partial), await it
            if inspect.iscoroutine(result):
                await result
    except Exception as e:
        logger.warning(f"Callback execution failed: {e}", exc_info=True)


def launch_callback_task(
    callback: Callable | None,
    data: Any,  # noqa: ANN401
    unpack: bool = False,
) -> None:
    """Launch a callback as a background task (fire-and-forget).

    Args:
        callback: The callback callable to execute (can be sync or async)
        data: The data to pass to the callback
        unpack: If True and data is a dict, unpack it as keyword arguments

    Note:
        This launches the callback without blocking. Errors are logged but won't
        affect the main execution flow.
    """
    if callback is None:
        return

    async def _run_callback() -> None:
        """Wrapper to run callback with error handling."""
        try:
            if inspect.iscoroutinefunction(callback):
                if unpack and isinstance(data, dict):
                    # Filter parameters to match callback signature
                    filtered_data = filter_callback_parameters(callback, data)
                    await callback(**filtered_data)
                else:
                    await callback(data)
            else:
                if unpack and isinstance(data, dict):
                    # Filter parameters to match callback signature
                    filtered_data = filter_callback_parameters(callback, data)
                    result = callback(**filtered_data)
                else:
                    result = callback(data)
                if inspect.iscoroutine(result):
                    await result
        except Exception as e:
            logger.warning(f"Callback execution failed: {e}", exc_info=True)

    # Create task without awaiting (fire-and-forget)
    try:
        asyncio.create_task(_run_callback())
    except RuntimeError:
        # If no event loop is running, log a warning
        logger.warning("Cannot launch callback task: no event loop running")


def execute_combined_callback(
    callback: Callable | None,
    input_params: dict[str, Any],
    mock_output: Any,  # noqa: ANN401
) -> None:
    """Execute a combined callback synchronously with input parameters and mock output.

    Args:
        callback: The callback callable to execute
        input_params: Dictionary of input parameters
        mock_output: The output from the mock/stub call

    Note:
        Exceptions in callbacks are caught and logged to prevent breaking the main flow.
    """
    if callback is None:
        return

    try:
        # Combine input params with mock_output
        combined_data = {**input_params, "mock_output": mock_output}
        # Filter parameters to match callback signature
        filtered_data = filter_callback_parameters(callback, combined_data)
        callback(**filtered_data)
    except Exception as e:
        logger.warning(f"Combined callback execution failed: {e}", exc_info=True)


def launch_combined_callback_task(
    callback: Callable | None,
    input_params: dict[str, Any],
    mock_output: Any,  # noqa: ANN401
) -> None:
    """Launch a combined callback as a background task (fire-and-forget).

    Args:
        callback: The callback callable to execute (can be sync or async)
        input_params: Dictionary of input parameters
        mock_output: The output from the mock/stub call

    Note:
        This launches the callback without blocking. Errors are logged but won't
        affect the main execution flow.
    """
    if callback is None:
        return

    async def _run_callback() -> None:
        """Wrapper to run combined callback with error handling."""
        try:
            # Combine input params with mock_output
            combined_data = {**input_params, "mock_output": mock_output}
            # Filter parameters to match callback signature
            filtered_data = filter_callback_parameters(callback, combined_data)

            if inspect.iscoroutinefunction(callback):
                await callback(**filtered_data)
            else:
                result = callback(**filtered_data)
                if inspect.iscoroutine(result):
                    await result
        except Exception as e:
            logger.warning(f"Combined callback execution failed: {e}", exc_info=True)

    # Create task without awaiting (fire-and-forget)
    try:
        asyncio.create_task(_run_callback())
    except RuntimeError:
        # If no event loop is running, log a warning
        logger.warning("Cannot launch combined callback task: no event loop running")



"""Centralized API client for VERIS simulation endpoints."""

import logging
import os
from contextvars import ContextVar
from typing import Any

import httpx
from urllib.parse import urljoin

logger = logging.getLogger(__name__)

# Context variable to store base_url for per-request multi-tenant scenarios
_base_url_context: ContextVar[str | None] = ContextVar("veris_base_url", default=None)


class SimulatorAPIClient:
    """Centralized client for making requests to VERIS simulation endpoints.

    Note:
        This client intentionally reads configuration (base URL and API key)
        from environment variables at call-time instead of at construction
        time. This allows tests to patch environment variables and have those
        changes reflected without recreating the singleton.
    """

    def __init__(self, timeout: float | None = None) -> None:
        """Initialize the API client with static timeout configuration."""
        self._timeout = timeout or float(os.getenv("VERIS_MOCK_TIMEOUT", "300.0"))

    @property
    def base_url(self) -> str:
        """Get the resolved base URL."""
        return self._get_base_url()

    def _get_base_url(self) -> str:
        """Resolve the base URL from context, environment, or default.

        Priority order:
            1. Context variable (for per-request multi-tenant scenarios)
            2. VERIS_API_URL environment variable
            3. Default production URL

        Behavior:
            - If VERIS_API_URL is unset, default to the dev simulator URL.
            - If VERIS_API_URL is set to an empty string, treat it as empty
              (do not fall back). This supports tests expecting connection
              failures when an invalid endpoint is provided.
        """
        # Priority: context > env var > default
        context_url = _base_url_context.get()
        return context_url or os.getenv("VERIS_API_URL") or "https://simulator.api.veris.ai"

    def _build_headers(self) -> dict[str, str] | None:
        """Build headers including OpenTelemetry tracing and API key."""
        headers: dict[str, str] | None = None
        # Add API key header if available
        api_key = os.getenv("VERIS_API_KEY")
        if api_key:
            if headers is None:
                headers = {}
            headers["x-api-key"] = api_key

        return headers

    def post(self, endpoint: str, payload: dict[str, Any]) -> Any:  # noqa: ANN401
        """Make a synchronous POST request to the specified endpoint."""
        headers = self._build_headers()
        # Validate endpoint URL; raise ConnectError for non-absolute URLs to
        # mirror connection failures in tests when base URL is intentionally invalid.
        if not endpoint.startswith(("http://", "https://")):
            raise httpx.ConnectError("Invalid endpoint URL (not absolute): {endpoint}")

        with httpx.Client(timeout=self._timeout) as client:
            response = client.post(endpoint, json=payload, headers=headers)
            response.raise_for_status()
            return response.json() if response.content else None

    async def post_async(self, endpoint: str, payload: dict[str, Any]) -> Any:  # noqa: ANN401
        """Make an asynchronous POST request to the specified endpoint.

        This method uses httpx.AsyncClient and is safe to call from async functions
        without blocking the event loop.
        """
        headers = self._build_headers()
        # Validate endpoint URL; raise ConnectError for non-absolute URLs to
        # mirror connection failures in tests when base URL is intentionally invalid.
        if not endpoint.startswith(("http://", "https://")):
            error_msg = f"Invalid endpoint URL (not absolute): {endpoint}"
            raise httpx.ConnectError(error_msg)

        async with httpx.AsyncClient(timeout=self._timeout) as client:
            response = await client.post(endpoint, json=payload, headers=headers)
            response.raise_for_status()
            return response.json() if response.content else None

    @property
    def tool_mock_endpoint(self) -> str:
        """Get the tool mock endpoint URL."""
        return urljoin(self.base_url, "v3/tool_mock")

    def get_log_tool_call_endpoint(self, session_id: str) -> str:
        """Get the log tool call endpoint URL."""
        return urljoin(self.base_url, f"v3/log_tool_call?session_id={session_id}")

    def get_log_tool_response_endpoint(self, session_id: str) -> str:
        """Get the log tool response endpoint URL."""
        return urljoin(self.base_url, f"v3/log_tool_response?session_id={session_id}")


# Global singleton instance
_api_client = SimulatorAPIClient()


def get_api_client() -> SimulatorAPIClient:
    """Get the global API client instance."""
    return _api_client


def set_api_client_params(timeout: float | None = None) -> None:
    """Set the global API client instance for testing purposes."""
    global _api_client  # noqa: PLW0603
    _api_client = SimulatorAPIClient(timeout=timeout)



"""Tests for input_side_effect and output_side_effect functionality."""

import asyncio
from typing import Any
from unittest.mock import Mock, patch

import pytest

from veris_ai import veris
from veris_ai.api_client import get_api_client


def create_test_token(
    session_id: str = "test-session", thread_id: str | None = "test-thread"
) -> str:
    """Create a base64-encoded test token."""
    import base64
    import json

    token_data = {"session_id": session_id}
    if thread_id is not None:
        token_data["thread_id"] = thread_id
    return base64.b64encode(json.dumps(token_data).encode("utf-8")).decode("utf-8")


@pytest.fixture
def simulation_env():
    """Set up simulation environment with session_id."""
    token = create_test_token("test-session", "test-thread")
    veris.parse_token(token)
    yield
    veris.clear_context()


@pytest.fixture
def production_env():
    """Set up production environment without session_id."""
    veris.clear_context()
    yield


# Test mock decorator with input_side_effect


@pytest.mark.asyncio
async def test_mock_with_async_input_side_effect(simulation_env):
    """Test mock decorator with async input_side_effect."""
    captured_inputs = []

    async def capture_input(**kwargs):
        captured_inputs.append(kwargs)

    @veris.mock(mode="function", input_callback=capture_input)
    async def test_func(x: int, y: str) -> dict:
        return {"real": True}

    mock_response = {"mocked": True}

    with patch.object(get_api_client(), "post_async", return_value=mock_response):
        result = await test_func(42, "hello")
        assert result == {"mocked": True}
        # Wait for background tasks to complete
        await asyncio.sleep(0.01)
        assert len(captured_inputs) == 1
        assert captured_inputs[0] == {"x": 42, "y": "hello"}


@pytest.mark.asyncio
async def test_mock_with_sync_input_side_effect_async_func(simulation_env):
    """Test mock decorator with sync input_side_effect on async function."""
    captured_inputs = []

    def capture_input(**kwargs):
        captured_inputs.append(kwargs)

    @veris.mock(mode="function", input_callback=capture_input)
    async def test_func(a: int, b: int) -> int:
        return a + b

    mock_response = 100

    with patch.object(get_api_client(), "post_async", return_value=mock_response):
        result = await test_func(10, 20)
        assert result == 100
        # Wait for background tasks to complete
        await asyncio.sleep(0.01)
        assert len(captured_inputs) == 1
        assert captured_inputs[0] == {"a": 10, "b": 20}


def test_mock_with_sync_input_side_effect_sync_func(simulation_env):
    """Test mock decorator with sync input_side_effect on sync function."""
    captured_inputs = []

    def capture_input(**kwargs):
        captured_inputs.append(kwargs)

    @veris.mock(mode="function", input_callback=capture_input)
    def test_func(name: str, count: int) -> dict:
        return {"name": name, "count": count}

    mock_response = {"mocked": "data"}

    with patch.object(get_api_client(), "post", return_value=mock_response):
        result = test_func("test", 5)
        assert result == {"mocked": "data"}
        assert len(captured_inputs) == 1
        assert captured_inputs[0] == {"name": "test", "count": 5}


# Test mock decorator with output_side_effect


@pytest.mark.asyncio
async def test_mock_with_async_output_side_effect(simulation_env):
    """Test mock decorator with async output_side_effect."""
    captured_outputs = []

    async def capture_output(result: Any):
        captured_outputs.append(result)

    @veris.mock(mode="function", output_callback=capture_output)
    async def test_func(x: int) -> dict:
        return {"real": x}

    mock_response = {"mocked": 999}

    with patch.object(get_api_client(), "post_async", return_value=mock_response):
        result = await test_func(42)
        assert result == {"mocked": 999}
        # Wait for background tasks to complete
        await asyncio.sleep(0.01)
        assert len(captured_outputs) == 1
        assert captured_outputs[0] == {"mocked": 999}


@pytest.mark.asyncio
async def test_mock_with_sync_output_side_effect_async_func(simulation_env):
    """Test mock decorator with sync output_side_effect on async function."""
    captured_outputs = []

    def capture_output(result: Any):
        captured_outputs.append(result)

    @veris.mock(mode="function", output_callback=capture_output)
    async def test_func() -> str:
        return "real"

    mock_response = "mocked_string"

    with patch.object(get_api_client(), "post_async", return_value=mock_response):
        result = await test_func()
        assert result == "mocked_string"
        # Wait for background tasks to complete
        await asyncio.sleep(0.01)
        assert len(captured_outputs) == 1
        assert captured_outputs[0] == "mocked_string"


def test_mock_with_sync_output_side_effect_sync_func(simulation_env):
    """Test mock decorator with sync output_side_effect on sync function."""
    captured_outputs = []

    def capture_output(result: Any):
        captured_outputs.append(result)

    @veris.mock(mode="function", output_callback=capture_output)
    def test_func(x: int) -> int:
        return x * 2

    mock_response = 1000

    with patch.object(get_api_client(), "post", return_value=mock_response):
        result = test_func(10)
        assert result == 1000
        assert len(captured_outputs) == 1
        assert captured_outputs[0] == 1000


# Test mock decorator with both side effects


@pytest.mark.asyncio
async def test_mock_with_both_side_effects(simulation_env):
    """Test mock decorator with both input and output side effects."""
    captured_inputs = []
    captured_outputs = []

    async def capture_input(**kwargs):
        captured_inputs.append(kwargs)

    async def capture_output(result: Any):
        captured_outputs.append(result)

    @veris.mock(mode="function", input_callback=capture_input, output_callback=capture_output)
    async def test_func(x: int, y: str) -> dict:
        return {"real": True}

    mock_response = {"mocked": "result"}

    with patch.object(get_api_client(), "post_async", return_value=mock_response):
        result = await test_func(100, "test")
        assert result == {"mocked": "result"}
        # Wait for background tasks to complete
        await asyncio.sleep(0.01)
        assert len(captured_inputs) == 1
        assert captured_inputs[0] == {"x": 100, "y": "test"}
        assert len(captured_outputs) == 1
        assert captured_outputs[0] == {"mocked": "result"}


# Test stub decorator with input_side_effect


@pytest.mark.asyncio
async def test_stub_with_async_input_side_effect(simulation_env):
    """Test stub decorator with async input_side_effect."""
    captured_inputs = []

    async def capture_input(**kwargs):
        captured_inputs.append(kwargs)

    @veris.stub(return_value={"stubbed": True}, input_callback=capture_input)
    async def test_func(a: str, b: int) -> dict:
        return {"real": True}

    result = await test_func("hello", 42)
    assert result == {"stubbed": True}
    # Wait for background tasks to complete
    await asyncio.sleep(0.01)
    assert len(captured_inputs) == 1
    assert captured_inputs[0] == {"a": "hello", "b": 42}


@pytest.mark.asyncio
async def test_stub_with_sync_input_side_effect_async_func(simulation_env):
    """Test stub decorator with sync input_side_effect on async function."""
    captured_inputs = []

    def capture_input(**kwargs):
        captured_inputs.append(kwargs)

    @veris.stub(return_value=99, input_callback=capture_input)
    async def test_func(x: int) -> int:
        return x * 2

    result = await test_func(5)
    assert result == 99
    # Wait for background tasks to complete
    await asyncio.sleep(0.01)
    assert len(captured_inputs) == 1
    assert captured_inputs[0] == {"x": 5}


def test_stub_with_sync_input_side_effect_sync_func(simulation_env):
    """Test stub decorator with sync input_side_effect on sync function."""
    captured_inputs = []

    def capture_input(**kwargs):
        captured_inputs.append(kwargs)

    @veris.stub(return_value="stubbed", input_callback=capture_input)
    def test_func(name: str, value: float) -> str:
        return f"{name}: {value}"

    result = test_func("test", 3.14)
    assert result == "stubbed"
    assert len(captured_inputs) == 1
    assert captured_inputs[0] == {"name": "test", "value": 3.14}


# Test stub decorator with output_side_effect


@pytest.mark.asyncio
async def test_stub_with_async_output_side_effect(simulation_env):
    """Test stub decorator with async output_side_effect."""
    captured_outputs = []

    async def capture_output(result: Any):
        captured_outputs.append(result)

    @veris.stub(return_value=["stubbed", "list"], output_callback=capture_output)
    async def test_func() -> list:
        return ["real", "list"]

    result = await test_func()
    assert result == ["stubbed", "list"]
    # Wait for background tasks to complete
    await asyncio.sleep(0.01)
    assert len(captured_outputs) == 1
    assert captured_outputs[0] == ["stubbed", "list"]


@pytest.mark.asyncio
async def test_stub_with_sync_output_side_effect_async_func(simulation_env):
    """Test stub decorator with sync output_side_effect on async function."""
    captured_outputs = []

    def capture_output(result: Any):
        captured_outputs.append(result)

    @veris.stub(return_value={"stub": "data"}, output_callback=capture_output)
    async def test_func(x: int) -> dict:
        return {"real": x}

    result = await test_func(10)
    assert result == {"stub": "data"}
    # Wait for background tasks to complete
    await asyncio.sleep(0.01)
    assert len(captured_outputs) == 1
    assert captured_outputs[0] == {"stub": "data"}


def test_stub_with_sync_output_side_effect_sync_func(simulation_env):
    """Test stub decorator with sync output_side_effect on sync function."""
    captured_outputs = []

    def capture_output(result: Any):
        captured_outputs.append(result)

    @veris.stub(return_value=42, output_callback=capture_output)
    def test_func() -> int:
        return 100

    result = test_func()
    assert result == 42
    assert len(captured_outputs) == 1
    assert captured_outputs[0] == 42


# Test stub decorator with both side effects


@pytest.mark.asyncio
async def test_stub_with_both_side_effects(simulation_env):
    """Test stub decorator with both input and output side effects."""
    captured_inputs = []
    captured_outputs = []

    async def capture_input(**kwargs):
        captured_inputs.append(kwargs)

    async def capture_output(result: Any):
        captured_outputs.append(result)

    @veris.stub(
        return_value={"stubbed": "response"},
        input_callback=capture_input,
        output_callback=capture_output,
    )
    async def test_func(x: int, y: str) -> dict:
        return {"real": "response"}

    result = await test_func(123, "test_value")
    assert result == {"stubbed": "response"}
    # Wait for background tasks to complete
    await asyncio.sleep(0.01)
    assert len(captured_inputs) == 1
    assert captured_inputs[0] == {"x": 123, "y": "test_value"}
    assert len(captured_outputs) == 1
    assert captured_outputs[0] == {"stubbed": "response"}


# Test that ctx parameter is preserved


@pytest.mark.asyncio
async def test_mock_preserves_ctx_parameter(simulation_env):
    """Test that ctx parameter is preserved in input_side_effect."""
    captured_inputs = []

    async def capture_input(**kwargs):
        captured_inputs.append(kwargs)

    @veris.mock(mode="function", input_callback=capture_input)
    async def test_func(x: int, ctx: dict) -> dict:
        return {"x": x, "ctx": ctx}

    mock_response = {"mocked": True}
    context = {"session": "abc123", "user": "test_user"}

    with patch.object(get_api_client(), "post_async", return_value=mock_response):
        result = await test_func(42, ctx=context)
        assert result == {"mocked": True}
        # Wait for background tasks to complete
        await asyncio.sleep(0.01)
        assert len(captured_inputs) == 1
        # ctx should be preserved in input parameters
        assert "ctx" in captured_inputs[0]
        assert captured_inputs[0]["ctx"] == context
        assert captured_inputs[0]["x"] == 42


@pytest.mark.asyncio
async def test_stub_preserves_ctx_parameter(simulation_env):
    """Test that ctx parameter is preserved in input_side_effect for stub."""
    captured_inputs = []

    def capture_input(**kwargs):
        captured_inputs.append(kwargs)

    @veris.stub(return_value="stubbed", input_callback=capture_input)
    async def test_func(name: str, ctx: dict) -> str:
        return f"{name} with {ctx}"

    context = {"request_id": "12345"}
    result = await test_func("test", ctx=context)
    assert result == "stubbed"
    # Wait for background tasks to complete
    await asyncio.sleep(0.01)
    assert len(captured_inputs) == 1
    assert "ctx" in captured_inputs[0]
    assert captured_inputs[0]["ctx"] == context
    assert captured_inputs[0]["name"] == "test"


# Test that self and cls are excluded


@pytest.mark.asyncio
async def test_mock_excludes_self_parameter(simulation_env):
    """Test that self parameter is excluded from input_side_effect."""
    captured_inputs = []

    async def capture_input(**kwargs):
        captured_inputs.append(kwargs)

    class TestClass:
        @veris.mock(mode="function", input_callback=capture_input)
        async def test_method(self, x: int) -> dict:
            return {"x": x}

    mock_response = {"mocked": True}

    with patch.object(get_api_client(), "post_async", return_value=mock_response):
        obj = TestClass()
        result = await obj.test_method(42)
        assert result == {"mocked": True}
        # Wait for background tasks to complete
        await asyncio.sleep(0.01)
        assert len(captured_inputs) == 1
        # self should NOT be in input parameters
        assert "self" not in captured_inputs[0]
        assert captured_inputs[0] == {"x": 42}


# Test side effects in production mode (should not be called)


@pytest.mark.asyncio
async def test_mock_side_effects_not_called_in_production_mode(production_env):
    """Test that side effects are not called in production mode."""
    input_mock = Mock()
    output_mock = Mock()

    @veris.mock(mode="function", input_callback=input_mock, output_callback=output_mock)
    async def test_func(x: int) -> int:
        return x * 2

    result = await test_func(21)
    assert result == 42

    # Side effects should NOT have been called
    input_mock.assert_not_called()
    output_mock.assert_not_called()


@pytest.mark.asyncio
async def test_stub_side_effects_not_called_in_production_mode(production_env):
    """Test that side effects are not called in production mode for stub."""
    input_mock = Mock()
    output_mock = Mock()

    @veris.stub(return_value="stubbed", input_callback=input_mock, output_callback=output_mock)
    async def test_func(x: int) -> str:
        return f"real: {x}"

    result = await test_func(10)
    assert result == "real: 10"

    # Side effects should NOT have been called
    input_mock.assert_not_called()
    output_mock.assert_not_called()


# Test side effect error handling


@pytest.mark.asyncio
async def test_mock_input_side_effect_exception_is_logged(simulation_env):
    """Test that exceptions in input side effects are caught and logged."""

    async def failing_side_effect(**kwargs):
        raise ValueError("Side effect error")

    @veris.mock(mode="function", input_callback=failing_side_effect)
    async def test_func(x: int) -> dict:
        return {"x": x}

    mock_response = {"mocked": True}

    with (
        patch.object(get_api_client(), "post_async", return_value=mock_response),
        patch("veris_ai.utils.logger.warning") as mock_logger,
    ):
        # Function should still work despite side effect failure
        result = await test_func(42)
        assert result == {"mocked": True}

        # Wait for background tasks to complete
        await asyncio.sleep(0.01)

        # Error should have been logged
        mock_logger.assert_called_once()
        assert "Callback execution failed" in str(mock_logger.call_args)


@pytest.mark.asyncio
async def test_mock_output_side_effect_exception_is_logged(simulation_env):
    """Test that exceptions in output side effects are caught and logged."""

    def failing_side_effect(result: Any):
        raise RuntimeError("Output side effect error")

    @veris.mock(mode="function", output_callback=failing_side_effect)
    async def test_func() -> dict:
        return {"x": 1}

    mock_response = {"mocked": "response"}

    with (
        patch.object(get_api_client(), "post_async", return_value=mock_response),
        patch("veris_ai.utils.logger.warning") as mock_logger,
    ):
        # Function should still work and return result despite side effect failure
        result = await test_func()
        assert result == {"mocked": "response"}

        # Wait for background tasks to complete
        await asyncio.sleep(0.01)

        # Error should have been logged
        mock_logger.assert_called_once()
        assert "Callback execution failed" in str(mock_logger.call_args)


def test_stub_input_side_effect_exception_is_logged_sync(simulation_env):
    """Test that exceptions in sync stub input side effects are caught and logged."""

    def failing_side_effect(**kwargs):
        raise TypeError("Sync side effect error")

    @veris.stub(return_value=42, input_callback=failing_side_effect)
    def test_func(x: int) -> int:
        return x * 2

    with patch("veris_ai.utils.logger.warning") as mock_logger:
        # Function should still work despite side effect failure
        result = test_func(10)
        assert result == 42

        # Error should have been logged
        mock_logger.assert_called_once()
        assert "Callback execution failed" in str(mock_logger.call_args)


# Test that side effects can perform operations on parameters


@pytest.mark.asyncio
async def test_mock_input_side_effect_can_process_parameters(simulation_env):
    """Test that input side effect can perform operations on parameters."""
    computation_results = []

    async def compute_sum(**kwargs):
        """Side effect that performs arithmetic on input params."""
        result = kwargs["a"] + kwargs["b"]
        computation_results.append(result)

    @veris.mock(mode="function", input_callback=compute_sum)
    async def add_numbers(a: int, b: int) -> int:
        """Original function adds numbers."""
        return a + b

    mock_response = 999  # Mocked return value

    with patch.object(get_api_client(), "post_async", return_value=mock_response):
        result = await add_numbers(10, 20)
        # Function returns mocked value
        assert result == 999
        # Wait for background task
        await asyncio.sleep(0.01)
        # Side effect computed the actual sum of inputs
        assert len(computation_results) == 1
        assert computation_results[0] == 30  # 10 + 20


@pytest.mark.asyncio
async def test_mock_output_side_effect_can_process_result(simulation_env):
    """Test that output side effect can perform operations on result."""
    transformed_results = []

    async def transform_output(result: dict):
        """Side effect that transforms the output."""
        transformed = {k.upper(): v * 2 for k, v in result.items()}
        transformed_results.append(transformed)

    @veris.mock(mode="function", output_callback=transform_output)
    async def get_data(x: int) -> dict:
        return {"value": x}

    mock_response = {"value": 5, "count": 3}

    with patch.object(get_api_client(), "post_async", return_value=mock_response):
        result = await get_data(100)
        # Function returns original mocked value
        assert result == {"value": 5, "count": 3}
        # Wait for background task
        await asyncio.sleep(0.01)
        # Side effect transformed the output
        assert len(transformed_results) == 1
        assert transformed_results[0] == {"VALUE": 10, "COUNT": 6}


def test_stub_input_side_effect_can_validate_parameters(simulation_env):
    """Test that input side effect can validate parameters."""
    validation_results = []

    def validate_inputs(**kwargs):
        """Side effect that validates input parameters."""
        is_valid = kwargs["age"] >= 0 and len(kwargs["name"]) > 0
        validation_results.append(is_valid)

    @veris.stub(return_value="processed", input_callback=validate_inputs)
    def process_user(name: str, age: int) -> str:
        return f"User: {name}, Age: {age}"

    result = process_user("Alice", 25)
    # Function returns stubbed value
    assert result == "processed"
    # Side effect validated the inputs
    assert len(validation_results) == 1
    assert validation_results[0] is True


def test_mock_output_side_effect_divides_int_result(simulation_env):
    """Test output side effect that processes int result (divide by 2 case)."""
    divided_results = []

    def divide_by_2(result: int):
        """Side effect that divides the int result by 2."""
        divided = result / 2
        divided_results.append(divided)

    @veris.mock(mode="function", output_callback=divide_by_2)
    def add(a: int, b: int) -> int:
        """Add two numbers."""
        return a + b

    mock_response = 100  # Mocked return value

    with patch.object(get_api_client(), "post", return_value=mock_response):
        # Call the function
        result = add(10, 20)

        # Function returns the mocked value (100, not 30)
        assert result == 100

        # Side effect divided the mocked result by 2
        assert len(divided_results) == 1
        assert divided_results[0] == 50.0  # 100 / 2


@pytest.mark.asyncio
async def test_mock_output_side_effect_processes_int_async(simulation_env):
    """Test async output side effect that processes int result."""
    processed_results = []

    async def divide_by_2(result: int):
        """Async side effect that divides the int result by 2."""
        divided = result / 2
        processed_results.append(divided)

    @veris.mock(mode="function", output_callback=divide_by_2)
    async def add(a: int, b: int) -> int:
        """Add two numbers."""
        return a + b

    mock_response = 80  # Mocked return value

    with patch.object(get_api_client(), "post_async", return_value=mock_response):
        # Call the function
        result = await add(15, 25)

        # Function returns the mocked value (80, not 40)
        assert result == 80

        # Wait for background task to complete
        await asyncio.sleep(0.01)

        # Side effect divided the mocked result by 2
        assert len(processed_results) == 1
        assert processed_results[0] == 40.0  # 80 / 2


@pytest.mark.asyncio
async def test_mock_side_effect_with_string_operations(simulation_env):
    """Test side effect performing string operations on inputs."""
    formatted_logs = []

    async def log_formatted(**kwargs):
        """Side effect that formats parameters as a log message."""
        log_msg = f"Calling multiply with x={kwargs['x']}, y={kwargs['y']}"
        formatted_logs.append(log_msg)

    @veris.mock(mode="function", input_callback=log_formatted)
    async def multiply(x: int, y: int) -> int:
        return x * y

    mock_response = 42

    with patch.object(get_api_client(), "post_async", return_value=mock_response):
        result = await multiply(7, 8)
        assert result == 42
        await asyncio.sleep(0.01)
        assert len(formatted_logs) == 1
        assert formatted_logs[0] == "Calling multiply with x=7, y=8"


@pytest.mark.asyncio
async def test_stub_side_effect_with_complex_computation(simulation_env):
    """Test side effect performing complex computation on both inputs and outputs."""
    metrics = []

    async def record_metrics(**kwargs):
        """Side effect that computes metrics from inputs."""
        total = sum(kwargs["values"])
        avg = total / len(kwargs["values"]) if kwargs["values"] else 0
        metrics.append({"total": total, "avg": avg, "count": len(kwargs["values"])})

    @veris.stub(return_value={"status": "ok"}, input_callback=record_metrics)
    async def process_values(values: list[int]) -> dict:
        return {"result": sum(values)}

    result = await process_values([10, 20, 30, 40])
    assert result == {"status": "ok"}
    await asyncio.sleep(0.01)
    assert len(metrics) == 1
    assert metrics[0]["total"] == 100
    assert metrics[0]["avg"] == 25.0
    assert metrics[0]["count"] == 4


def test_mock_side_effect_end_to_end_with_real_function_signature(simulation_env):
    """Test complete end-to-end flow with realistic function."""
    # Track what the side effect computed
    computed_differences = []

    def compute_difference(**kwargs):
        """Side effect computes difference of inputs."""
        diff = kwargs["a"] - kwargs["b"]
        computed_differences.append(diff)

    @veris.mock(mode="function", input_callback=compute_difference)
    def add_numbers(a: int, b: int) -> int:
        """This function adds two numbers."""
        return a + b

    mock_response = 999

    with patch.object(get_api_client(), "post", return_value=mock_response):
        # Call the decorated function
        result = add_numbers(50, 30)

        # Function returns mocked value (not the real a+b)
        assert result == 999

        # Side effect computed a-b correctly
        assert len(computed_differences) == 1
        assert computed_differences[0] == 20  # 50 - 30


@pytest.mark.asyncio
async def test_mock_side_effects_with_ctx_parameter_operations(simulation_env):
    """Test that side effect can access and use ctx parameter."""
    ctx_metadata = []

    async def extract_metadata(**kwargs):
        """Side effect extracts metadata from ctx."""
        ctx = kwargs.get("ctx", {})
        metadata = {
            "user": ctx.get("user"),
            "session": ctx.get("session"),
            "input_value": kwargs.get("value"),
        }
        ctx_metadata.append(metadata)

    @veris.mock(mode="function", input_callback=extract_metadata)
    async def process_data(value: int, ctx: dict) -> dict:
        return {"result": value * 2}

    mock_response = {"mocked": True}
    context = {"user": "alice", "session": "xyz789"}

    with patch.object(get_api_client(), "post_async", return_value=mock_response):
        result = await process_data(42, ctx=context)
        assert result == {"mocked": True}
        await asyncio.sleep(0.01)
        assert len(ctx_metadata) == 1
        assert ctx_metadata[0] == {
            "user": "alice",
            "session": "xyz789",
            "input_value": 42,
        }


# Test execution order


@pytest.mark.asyncio
async def test_mock_side_effects_execution_order(simulation_env):
    """Test that side effects execute after mock call as background tasks."""
    execution_order = []

    async def input_effect(**kwargs):
        execution_order.append("input")

    async def output_effect(result: Any):
        execution_order.append("output")

    @veris.mock(mode="function", input_callback=input_effect, output_callback=output_effect)
    async def test_func(x: int) -> dict:
        execution_order.append("original_func")
        return {"x": x}

    mock_response = {"mocked": True}

    with patch.object(get_api_client(), "post_async", return_value=mock_response):
        result = await test_func(42)
        assert result == {"mocked": True}
        # Wait for background tasks to complete
        await asyncio.sleep(0.01)
        # In simulation mode, original func is not called, side effects run as background tasks
        assert set(execution_order) == {"input", "output"}
        assert len(execution_order) == 2


@pytest.mark.asyncio
async def test_stub_side_effects_execution_order(simulation_env):
    """Test that side effects execute after stub as background tasks."""
    execution_order = []

    async def input_effect(**kwargs):
        execution_order.append("input")

    async def output_effect(result: Any):
        execution_order.append("output")

    @veris.stub(
        return_value={"stubbed": True},
        input_callback=input_effect,
        output_callback=output_effect,
    )
    async def test_func(x: int) -> dict:
        execution_order.append("original_func")
        return {"x": x}

    result = await test_func(42)
    assert result == {"stubbed": True}
    # Wait for background tasks to complete
    await asyncio.sleep(0.01)
    # In simulation mode, original func is not called, side effects run as background tasks
    assert set(execution_order) == {"input", "output"}
    assert len(execution_order) == 2


# Test callback signature compatibility


def test_callback_with_subset_of_parameters(simulation_env):
    """Test that callback with fewer parameters than the function works correctly."""
    multiplication_results = []

    def multiply(a: int, b: int, d: int = 4):
        """Callback that only takes a, b, and d (not c)."""
        result = a * b * d
        multiplication_results.append(result)

    @veris.mock(mode="function", input_callback=multiply)
    def add(a: int, b: int, c: int = None, d: int = None) -> int:
        """Function with more parameters than the callback accepts."""
        return a + b

    mock_response = 999

    with patch.object(get_api_client(), "post", return_value=mock_response):
        # Call with all parameters
        result = add(10, 20, c=5, d=2)
        assert result == 999

        # Callback should have been called with only a, b, d (not c)
        assert len(multiplication_results) == 1
        assert multiplication_results[0] == 10 * 20 * 2  # a * b * d = 400


@pytest.mark.asyncio
async def test_async_callback_with_subset_of_parameters(simulation_env):
    """Test that async callback with fewer parameters than the function works correctly."""
    processed_results = []

    async def process_subset(x: int, z: str = "default"):
        """Callback that only takes x and z (not y)."""
        processed_results.append(f"{x}:{z}")

    @veris.mock(mode="function", input_callback=process_subset)
    async def complex_func(x: int, y: float, z: str = "test") -> dict:
        """Function with more parameters than the callback accepts."""
        return {"result": x + y}

    mock_response = {"mocked": True}

    with patch.object(get_api_client(), "post_async", return_value=mock_response):
        # Call with all parameters
        result = await complex_func(42, 3.14, z="custom")
        assert result == {"mocked": True}

        # Wait for background task
        await asyncio.sleep(0.01)

        # Callback should have been called with only x and z (not y)
        assert len(processed_results) == 1
        assert processed_results[0] == "42:custom"


def test_callback_with_no_matching_parameters(simulation_env):
    """Test that callback with completely different parameters still works with **kwargs."""
    called = []

    def generic_callback(**kwargs):
        """Callback that accepts any parameters."""
        called.append(True)

    @veris.stub(return_value="result", input_callback=generic_callback)
    def some_func(a: int, b: str) -> str:
        return f"{a}:{b}"

    result = some_func(1, "test")
    assert result == "result"
    assert len(called) == 1


# Test combined_callback functionality


@pytest.mark.asyncio
async def test_mock_with_async_combined_callback(simulation_env):
    """Test mock decorator with async combined_callback."""
    captured_results = []

    async def combined_handler(a: int, b: int, d: int = None, mock_output: int = None):
        """Combined callback that uses both inputs and output."""
        # Compute sum of inputs that are not None, then multiply by mock_output
        total_input = a + b + (d if d is not None else 0)
        result = total_input * (mock_output if mock_output is not None else 1)
        captured_results.append(result)

    @veris.mock(mode="function", combined_callback=combined_handler)
    async def add(a: int, b: int, c: int = None, d: int = None) -> int:
        """Add numbers together."""
        return a + b + (c if c is not None else 0) + (d if d is not None else 0)

    mock_response = 6

    with patch.object(get_api_client(), "post_async", return_value=mock_response):
        result = await add(1, 2, d=4)
        assert result == 6
        # Wait for background tasks to complete
        await asyncio.sleep(0.01)
        assert len(captured_results) == 1
        # (1 + 2 + 4) * 6 = 42
        assert captured_results[0] == 42


@pytest.mark.asyncio
async def test_mock_with_sync_combined_callback_async_func(simulation_env):
    """Test mock decorator with sync combined_callback on async function."""
    captured_results = []

    def multiply_by_result(a: int, b: int, mock_output: int = None):
        """Multiply inputs by the mock output."""
        result = (a + b) * (mock_output if mock_output is not None else 1)
        captured_results.append(result)

    @veris.mock(mode="function", combined_callback=multiply_by_result)
    async def add(a: int, b: int) -> int:
        return a + b

    mock_response = 10

    with patch.object(get_api_client(), "post_async", return_value=mock_response):
        result = await add(3, 7)
        assert result == 10
        # Wait for background tasks to complete
        await asyncio.sleep(0.01)
        assert len(captured_results) == 1
        # (3 + 7) * 10 = 100
        assert captured_results[0] == 100


def test_mock_with_sync_combined_callback_sync_func(simulation_env):
    """Test mock decorator with sync combined_callback on sync function."""
    captured_results = []

    def process_combined(x: int, y: str, mock_output: dict = None):
        """Process both input and output."""
        result = {
            "input_x": x,
            "input_y": y,
            "output": mock_output,
            "combined": f"{y}:{x}:{mock_output.get('value') if mock_output else 'none'}",
        }
        captured_results.append(result)

    @veris.mock(mode="function", combined_callback=process_combined)
    def test_func(x: int, y: str) -> dict:
        return {"value": x}

    mock_response = {"value": 999}

    with patch.object(get_api_client(), "post", return_value=mock_response):
        result = test_func(42, "test")
        assert result == {"value": 999}
        assert len(captured_results) == 1
        assert captured_results[0]["input_x"] == 42
        assert captured_results[0]["input_y"] == "test"
        assert captured_results[0]["output"] == {"value": 999}
        assert captured_results[0]["combined"] == "test:42:999"


@pytest.mark.asyncio
async def test_stub_with_async_combined_callback(simulation_env):
    """Test stub decorator with async combined_callback."""
    captured_results = []

    async def combined_handler(name: str, count: int, mock_output: str = None):
        """Combine input and output."""
        result = f"{name}:{count}:{mock_output}"
        captured_results.append(result)

    @veris.stub(return_value="stubbed_value", combined_callback=combined_handler)
    async def process(name: str, count: int) -> str:
        return f"{name}_{count}"

    result = await process("test", 5)
    assert result == "stubbed_value"
    # Wait for background tasks to complete
    await asyncio.sleep(0.01)
    assert len(captured_results) == 1
    assert captured_results[0] == "test:5:stubbed_value"


def test_stub_with_sync_combined_callback_sync_func(simulation_env):
    """Test stub decorator with sync combined_callback on sync function."""
    captured_results = []

    def multiply_inputs_by_output(a: int, b: int, mock_output: int = None):
        """Multiply sum of inputs by output."""
        result = (a + b) * (mock_output if mock_output is not None else 1)
        captured_results.append(result)

    @veris.stub(return_value=5, combined_callback=multiply_inputs_by_output)
    def add(a: int, b: int) -> int:
        return a + b

    result = add(2, 3)
    assert result == 5
    assert len(captured_results) == 1
    # (2 + 3) * 5 = 25
    assert captured_results[0] == 25


@pytest.mark.asyncio
async def test_combined_callback_with_parameter_filtering(simulation_env):
    """Test that combined callback only receives parameters it accepts."""
    captured_results = []

    async def selective_callback(a: int, mock_output: int = None):
        """Callback that only accepts 'a' and 'mock_output', not 'b' or 'c'."""
        result = a * (mock_output if mock_output is not None else 1)
        captured_results.append(result)

    @veris.mock(mode="function", combined_callback=selective_callback)
    async def complex_func(a: int, b: int, c: str = "default") -> int:
        return a + b

    mock_response = 7

    with patch.object(get_api_client(), "post_async", return_value=mock_response):
        result = await complex_func(3, 5, c="ignored")
        assert result == 7
        # Wait for background tasks to complete
        await asyncio.sleep(0.01)
        assert len(captured_results) == 1
        # 3 * 7 = 21 (b and c are filtered out)
        assert captured_results[0] == 21


@pytest.mark.asyncio
async def test_combined_callback_in_production_mode(production_env):
    """Test that combined_callback is not called in production mode."""
    callback_mock = Mock()

    @veris.mock(mode="function", combined_callback=callback_mock)
    async def test_func(x: int) -> int:
        return x * 2

    result = await test_func(21)
    assert result == 42

    # Combined callback should NOT have been called
    callback_mock.assert_not_called()


@pytest.mark.asyncio
async def test_combined_callback_exception_is_logged(simulation_env):
    """Test that exceptions in combined callback are caught and logged."""

    async def failing_combined_callback(x: int, mock_output: int = None):
        raise ValueError("Combined callback error")

    @veris.mock(mode="function", combined_callback=failing_combined_callback)
    async def test_func(x: int) -> int:
        return x * 2

    mock_response = 100

    with (
        patch.object(get_api_client(), "post_async", return_value=mock_response),
        patch("veris_ai.utils.logger.warning") as mock_logger,
    ):
        # Function should still work despite combined callback failure
        result = await test_func(42)
        assert result == 100

        # Wait for background tasks to complete
        await asyncio.sleep(0.01)

        # Error should have been logged
        mock_logger.assert_called()
        assert "Combined callback execution failed" in str(mock_logger.call_args)


@pytest.mark.asyncio
async def test_mock_with_all_three_callbacks(simulation_env):
    """Test mock decorator with input, output, and combined callbacks."""
    captured_inputs = []
    captured_outputs = []
    captured_combined = []

    async def capture_input(**kwargs):
        captured_inputs.append(kwargs)

    async def capture_output(result: int):
        captured_outputs.append(result)

    async def capture_combined(a: int, b: int, mock_output: int = None):
        combined = {"sum": a + b, "product": a * b, "mock": mock_output}
        captured_combined.append(combined)

    @veris.mock(
        mode="function",
        input_callback=capture_input,
        output_callback=capture_output,
        combined_callback=capture_combined,
    )
    async def add(a: int, b: int) -> int:
        return a + b

    mock_response = 50

    with patch.object(get_api_client(), "post_async", return_value=mock_response):
        result = await add(10, 20)
        assert result == 50
        # Wait for background tasks to complete
        await asyncio.sleep(0.01)

        # All callbacks should have been called
        assert len(captured_inputs) == 1
        assert captured_inputs[0] == {"a": 10, "b": 20}

        assert len(captured_outputs) == 1
        assert captured_outputs[0] == 50

        assert len(captured_combined) == 1
        assert captured_combined[0] == {"sum": 30, "product": 200, "mock": 50}


def test_stub_with_all_three_callbacks(simulation_env):
    """Test stub decorator with input, output, and combined callbacks."""
    captured_inputs = []
    captured_outputs = []
    captured_combined = []

    def capture_input(**kwargs):
        captured_inputs.append(kwargs)

    def capture_output(result: str):
        captured_outputs.append(result)

    def capture_combined(name: str, value: int, mock_output: str = None):
        combined = f"{name}_{value}_{mock_output}"
        captured_combined.append(combined)

    @veris.stub(
        return_value="stubbed",
        input_callback=capture_input,
        output_callback=capture_output,
        combined_callback=capture_combined,
    )
    def process(name: str, value: int) -> str:
        return f"{name}:{value}"

    result = process("test", 42)
    assert result == "stubbed"

    # All callbacks should have been called
    assert len(captured_inputs) == 1
    assert captured_inputs[0] == {"name": "test", "value": 42}

    assert len(captured_outputs) == 1
    assert captured_outputs[0] == "stubbed"

    assert len(captured_combined) == 1
    assert captured_combined[0] == "test_42_stubbed"


@pytest.mark.asyncio
async def test_combined_callback_with_kwargs_accepts_all(simulation_env):
    """Test combined callback with **kwargs accepts all parameters including mock_output."""
    captured_data = []

    async def flexible_callback(**kwargs):
        """Callback that accepts any parameters via **kwargs."""
        captured_data.append(kwargs)

    @veris.mock(mode="function", combined_callback=flexible_callback)
    async def multi_param_func(a: int, b: str, c: float = 3.14) -> dict:
        return {"result": a}

    mock_response = {"result": 999}

    with patch.object(get_api_client(), "post_async", return_value=mock_response):
        result = await multi_param_func(10, "test", c=2.71)
        assert result == {"result": 999}
        # Wait for background tasks to complete
        await asyncio.sleep(0.01)

        assert len(captured_data) == 1
        # Should have all input params plus mock_output
        assert captured_data[0]["a"] == 10
        assert captured_data[0]["b"] == "test"
        assert captured_data[0]["c"] == 2.71
        assert captured_data[0]["mock_output"] == {"result": 999}


def test_combined_callback_user_example(simulation_env):
    """Test the exact example from the user: add function with multiply_by_result."""
    multiplication_results = []

    def multiply_by_result(a: int, b: int, d: int = None, mock_output: int = None):
        """Multiply sum of provided inputs by the mock output."""
        # Sum the inputs that are not None
        input_sum = a + b + (d if d is not None else 0)
        # Multiply by mock output
        result = input_sum * (mock_output if mock_output is not None else 1)
        multiplication_results.append(result)

    @veris.mock(mode="function", combined_callback=multiply_by_result)
    def add(a: int, b: int, c: int = None, d: int = None) -> int:
        """Add numbers together."""
        total = a + b
        if c is not None:
            total += c
        if d is not None:
            total += d
        return total

    # Mock the API to return 6 as the result
    mock_response = 6

    with patch.object(get_api_client(), "post", return_value=mock_response):
        # Call add(1, 2, d=4)
        # Expected: inputs are 1, 2, 4 (c is None)
        # Mock returns 6
        # Combined callback should compute: (1 + 2 + 4) * 6 = 42
        result = add(1, 2, d=4)

        # The function returns the mocked value
        assert result == 6

        # The combined callback computed (1+2+4)*6
        assert len(multiplication_results) == 1
        assert multiplication_results[0] == 42



"""OpenAI Agents wrapper for automatic tool mocking via Veris SDK."""

import inspect
import json
import logging
from collections.abc import Callable
from typing import Any

from agents import RunContextWrapper, RunResult, Runner as OpenAIRunner
from pydantic import BaseModel

from veris_ai import veris
from veris_ai.tool_mock import mock_tool_call, mock_tool_call_async
from veris_ai.models import ToolCallOptions

logger = logging.getLogger(__name__)


def _wrap(
    include_tools: list[str] | None = None,
    exclude_tools: list[str] | None = None,
    tool_options: dict[str, ToolCallOptions] | None = None,
) -> Callable:
    """Private wrapper for OpenAI agents Runner to intercept tool calls through Veris SDK.

    This function transparently intercepts tool calls from OpenAI agents and
    routes them through the Veris SDK's mocking infrastructure.

    Args:
        include_tools: Optional list of tool names to intercept (only these if provided)
        exclude_tools: Optional list of tool names to NOT intercept (these run normally)
        tool_options: Optional per-tool configuration for mocking behavior

    Returns:
        A wrapped Runner.run function

    Raises:
        ValueError: If both include_tools and exclude_tools are specified
        ImportError: If agents package is not installed
    """
    if include_tools and exclude_tools:
        msg = "Cannot specify both include_tools and exclude_tools"
        raise ValueError(msg)
    if not tool_options:
        tool_options = {}

    def wrapped_run_func(run_func: Callable) -> Callable:
        """Inner wrapper that takes the actual Runner.run function."""
        try:
            from agents import FunctionTool  # type: ignore[import-untyped] # noqa: PLC0415
        except ImportError as e:
            msg = "openai-agents package not installed. Install with: pip install veris-ai[agents]"
            raise ImportError(msg) from e

        async def wrapped_run(starting_agent: Any, input_text: str, **kwargs: Any) -> Any:  # noqa: ANN401
            """Wrapped version of Runner.run that intercepts tool calls."""
            # Store a mapping of tools to their original functions
            tool_functions = {}

            if hasattr(starting_agent, "tools") and starting_agent.tools:
                for tool in starting_agent.tools:
                    if isinstance(tool, FunctionTool):
                        tool_name = getattr(tool, "name", None)

                        # Check if we should patch this tool
                        if tool_name and _should_intercept_tool(
                            tool_name, include_tools, exclude_tools
                        ):
                            # Extract the original function before patching
                            original_func = _extract_the_func(tool.on_invoke_tool)
                            if original_func:
                                tool_functions[id(tool)] = original_func

                            # Store original on_invoke_tool
                            original_on_invoke = tool.on_invoke_tool

                            def make_wrapped_on_invoke_tool(
                                tool_id: int, orig_invoke: Callable, tool_name_inner: str
                            ) -> Callable:
                                """Create a wrapped on_invoke_tool with proper closure."""

                                async def wrapped_on_invoke_tool(
                                    ctx: RunContextWrapper[Any], parameters: str
                                ) -> Any:  # noqa: ANN401
                                    """Wrapped on_invoke_tool that intercepts the tool function."""
                                    session_id = veris.session_id
                                    thread_id = veris.thread_id
                                    the_func = tool_functions.get(tool_id)
                                    if the_func and session_id:
                                        # Check if async or sync, call appropriate version
                                        if inspect.iscoroutinefunction(the_func):
                                            # Use async version (non-blocking)
                                            return await mock_tool_call_async(
                                                the_func,
                                                session_id,
                                                json.loads(parameters),
                                                tool_options.get(tool_name_inner),
                                                thread_id=thread_id,
                                            )
                                        # Use sync version for sync functions
                                        return mock_tool_call(
                                            the_func,
                                            session_id,
                                            json.loads(parameters),
                                            tool_options.get(tool_name_inner),
                                            thread_id=thread_id,
                                        )
                                    # Fall back to original if we couldn't extract the function
                                    return await orig_invoke(ctx, parameters)

                                return wrapped_on_invoke_tool

                            tool.on_invoke_tool = make_wrapped_on_invoke_tool(
                                id(tool), original_on_invoke, tool_name
                            )
            return await run_func(starting_agent, input_text, **kwargs)

        return wrapped_run

    return wrapped_run_func


def _should_intercept_tool(
    tool_name: str,
    include_tools: list[str] | None,
    exclude_tools: list[str] | None,
) -> bool:
    """Determine if a tool should be intercepted based on include/exclude lists.

    Args:
        tool_name: Name of the tool
        include_tools: If provided, only these tools are intercepted
        exclude_tools: If provided, these tools are NOT intercepted

    Returns:
        True if the tool should be intercepted, False otherwise
    """
    if include_tools:
        return tool_name in include_tools
    if exclude_tools:
        return tool_name not in exclude_tools
    return True


def _extract_the_func(on_invoke_tool: Callable) -> Callable | None:
    """Extract the original user function from the on_invoke_tool closure.

    This function attempts multiple strategies to extract the original function:
    1. Direct attribute access (if the tool stores it)
    2. Closure inspection for known patterns
    3. Deep closure traversal as a fallback

    Args:
        on_invoke_tool: The on_invoke_tool function from FunctionTool

    Returns:
        The original user function if found, None otherwise
    """

    # Strategy 1: Check if the tool has stored the original function as an attribute
    # (This would be the cleanest approach if the agents library supported it)
    if hasattr(on_invoke_tool, "__wrapped__"):
        return on_invoke_tool.__wrapped__

    # Strategy 2: Look for the function in the closure using known structure
    # Based on the agents library implementation, we know:
    # - on_invoke_tool has _on_invoke_tool_impl in its closure
    # - _on_invoke_tool_impl has the_func in its closure

    if not hasattr(on_invoke_tool, "__closure__") or not on_invoke_tool.__closure__:
        return None

    # Find _on_invoke_tool_impl by looking for a function with that name pattern
    for cell in on_invoke_tool.__closure__:
        try:
            obj = cell.cell_contents
            if not callable(obj):
                continue

            # Check if this looks like _on_invoke_tool_impl
            if (
                hasattr(obj, "__name__")
                and "_on_invoke_tool_impl" in obj.__name__
                and hasattr(obj, "__closure__")
                and obj.__closure__
            ):
                # Now look for the_func in its closure
                return _find_user_function_in_closure(obj.__closure__)
        except (ValueError, AttributeError):
            continue

    # Strategy 3: Fallback - do a broader search in the closure
    return _find_user_function_in_closure(on_invoke_tool.__closure__)


def _find_user_function_in_closure(closure: tuple) -> Callable | None:
    """Find the user function in a closure by filtering out known library functions.

    Args:
        closure: The closure tuple to search

    Returns:
        The user function if found, None otherwise
    """
    # List of module prefixes that indicate library/framework code
    library_modules = ("json", "inspect", "agents", "pydantic", "openai", "typing")

    for cell in closure:
        try:
            obj = cell.cell_contents

            # Must be callable but not a class
            if not callable(obj) or isinstance(obj, type):
                continue

            # Skip private/internal functions
            if hasattr(obj, "__name__") and obj.__name__.startswith("_"):
                continue

            # Check the module to filter out library code
            module = inspect.getmodule(obj)
            if module:
                # Skip if it's from a known library
                if module.__name__.startswith(library_modules):
                    continue

                # Skip if it's from site-packages (library code)
                if (
                    hasattr(module, "__file__")
                    and module.__file__
                    and "site-packages" in module.__file__
                    # Unless it's user code installed as a package
                    # (this is a heuristic - may need adjustment)
                    and not any(pkg in module.__name__ for pkg in ["my_", "custom_", "app_"])
                ):
                    continue

            # If we made it here, this is likely the user function
            return obj

        except (ValueError, AttributeError):
            continue

    return None


class VerisConfig(BaseModel):
    """Configuration for the Veris SDK Runner.

    Attributes:
        include_tools: Optional list of tool names to intercept (only these if provided)
        exclude_tools: Optional list of tool names to NOT intercept (these run normally)
        tool_options: Optional per-tool configuration for mocking behavior
    """

    include_tools: list[str] | None = None
    exclude_tools: list[str] | None = None
    tool_options: dict[str, ToolCallOptions] | None = None


class Runner(OpenAIRunner):
    """Veris-enhanced Runner that extends OpenAI's Runner with tool interception.

    This class extends the OpenAI agents Runner to intercept tool calls
    and route them through the Veris SDK's mocking infrastructure.

    Example:
        ```python
        from veris_ai import Runner, VerisConfig
        from agents import Agent, FunctionTool

        # Define your agent with tools
        agent = Agent(...)

        # Use Veris Runner instead of OpenAI Runner
        result = await Runner.run(agent, "Process this input")

        # Or with configuration
        config = VerisConfig(include_tools=["calculator", "search"])
        result = await Runner.run(agent, "Calculate 2+2", veris_config=config)
        ```
    """

    @classmethod
    async def run(
        cls,
        starting_agent: Any,  # noqa: ANN401
        input: Any,  # noqa: ANN401, A002
        veris_config: VerisConfig | None = None,
        **kwargs: Any,  # noqa: ANN401
    ) -> RunResult:  # noqa: ANN401
        """Run an agent with Veris tool interception.

        This method overrides the OpenAI Runner.run to apply tool interception
        based on the provided configuration.

        Args:
            starting_agent: The OpenAI agent to run
            input: The input text/messages to process
            veris_config: Optional configuration for tool interception
            **kwargs: Additional arguments to pass to the base Runner.run

        Returns:
            The result from the agent execution
        """
        # Use provided config or create default
        config = veris_config or VerisConfig()

        # Validate configuration
        if config.include_tools and config.exclude_tools:
            msg = "Cannot specify both include_tools and exclude_tools"
            raise ValueError(msg)

        # Apply the wrapping logic
        wrapped_run = _wrap(
            include_tools=config.include_tools,
            exclude_tools=config.exclude_tools,
            tool_options=config.tool_options or {},
        )(OpenAIRunner.run)

        # Execute the wrapped run function
        return await wrapped_run(starting_agent, input, **kwargs)



"""Tests for SimulatorAPIClient endpoint URL generation and configuration."""

import os
from unittest.mock import patch

import pytest

from veris_ai.api_client import SimulatorAPIClient, get_api_client, set_api_client_params


def test_tool_mock_endpoint_default():
    """Test that tool_mock_endpoint property returns correct URL with default base_url."""
    client = SimulatorAPIClient()
    endpoint = client.tool_mock_endpoint

    # Should use default base URL
    assert endpoint == "https://simulator.api.veris.ai/v3/tool_mock"


def test_tool_mock_endpoint_with_env_var():
    """Test that tool_mock_endpoint respects VERIS_API_URL environment variable."""
    with patch.dict(os.environ, {"VERIS_API_URL": "https://test.api.veris.ai"}):
        client = SimulatorAPIClient()
        endpoint = client.tool_mock_endpoint

        assert endpoint == "https://test.api.veris.ai/v3/tool_mock"


def test_log_tool_call_endpoint():
    """Test get_log_tool_call_endpoint generates correct URL."""
    with patch.dict(os.environ, {"VERIS_API_URL": "https://test.api.com"}):
        client = SimulatorAPIClient()
        endpoint = client.get_log_tool_call_endpoint("session-123")

        assert endpoint == "https://test.api.com/v3/log_tool_call?session_id=session-123"


def test_log_tool_response_endpoint():
    """Test get_log_tool_response_endpoint generates correct URL."""
    with patch.dict(os.environ, {"VERIS_API_URL": "https://test.api.com"}):
        client = SimulatorAPIClient()
        endpoint = client.get_log_tool_response_endpoint("session-456")

        assert endpoint == "https://test.api.com/v3/log_tool_response?session_id=session-456"


def test_custom_timeout():
    """Test that custom timeout is set correctly."""
    client = SimulatorAPIClient(timeout=30.0)
    assert client._timeout == 30.0


def test_default_timeout():
    """Test that default timeout is used when not specified."""
    with patch.dict(os.environ, {"VERIS_MOCK_TIMEOUT": "120.0"}):
        client = SimulatorAPIClient()
        assert client._timeout == 120.0


def test_set_api_client_params_timeout():
    """Test that set_api_client_params reconfigures the global client timeout."""
    # Reconfigure with custom timeout
    set_api_client_params(timeout=60.0)

    # Get client and verify timeout has been updated
    client = get_api_client()
    assert client._timeout == 60.0

    # Restore original client
    set_api_client_params()


def test_empty_env_var_uses_default():
    """Test that empty VERIS_API_URL falls back to default."""
    with patch.dict(os.environ, {"VERIS_API_URL": ""}, clear=True):
        client = SimulatorAPIClient()
        # Empty string should NOT be used, should fall back to default
        assert client._get_base_url() == "https://simulator.api.veris.ai"



"""Veris AI Python SDK."""

from typing import Any

__version__ = "0.1.0"

# Import lightweight modules that only use base dependencies
from .jaeger_interface import JaegerClient
from .models import ResponseExpectation, ToolCallOptions
from .observability import init_observability, instrument_fastapi_app
from .tool_mock import veris

# Lazy import for modules with heavy dependencies
_Runner = None
_VerisConfig = None


def __getattr__(name: str) -> Any:  # noqa: ANN401
    """Lazy load Runner and VerisConfig classes."""
    global _Runner, _VerisConfig  # noqa: PLW0603
    if name == "Runner":
        if _Runner is None:
            try:
                from .agents_wrapper import Runner as _Runner_impl  # noqa: PLC0415

                _Runner = _Runner_impl
            except ImportError as e:
                error_msg = (
                    "The 'Runner' class requires additional dependencies. "
                    "Please install them with: pip install veris-ai[agents]"
                )
                raise ImportError(error_msg) from e
        return _Runner
    if name == "VerisConfig":
        if _VerisConfig is None:
            try:
                from .agents_wrapper import VerisConfig as _VerisConfig_impl  # noqa: PLC0415

                _VerisConfig = _VerisConfig_impl
            except ImportError as e:
                error_msg = (
                    "The 'VerisConfig' class requires additional dependencies. "
                    "Please install them with: pip install veris-ai[agents]"
                )
                raise ImportError(error_msg) from e
        return _VerisConfig
    msg = f"module {__name__!r} has no attribute {name!r}"
    raise AttributeError(msg)


__all__ = [
    "veris",
    "JaegerClient",
    "ResponseExpectation",
    "ToolCallOptions",
    "init_observability",
    "instrument_fastapi_app",
    "Runner",
    "VerisConfig",
]



"""Tests for veris_runner with ToolCallOptions configuration."""

import json
import os
import sys
import pytest
from typing import Any
from unittest.mock import patch, MagicMock

# Workaround for Python 3.13 httpx compatibility issue
if sys.version_info >= (3, 13):
    import httpx

    _ = httpx.AsyncClient

from agents import Agent, FunctionTool

from veris_ai import veris, Runner, VerisConfig, ToolCallOptions, ResponseExpectation
from tests.test_helpers import create_test_token

# Skip tests if no API key is available
if not os.environ.get("OPENAI_API_KEY"):
    pytest.skip(
        "OPENAI_API_KEY not set - skipping tests that require real API", allow_module_level=True
    )


# Tool implementations
async def calculator_impl(context: Any, arguments: str) -> str:
    """Calculator tool implementation."""
    args = json.loads(arguments)
    operation = args.get("operation", "add")
    a = args.get("a", 0)
    b = args.get("b", 0)

    if operation == "add":
        result = a + b
    elif operation == "subtract":
        result = a - b
    elif operation == "multiply":
        result = a * b
    elif operation == "divide":
        result = a / b if b != 0 else 0
    else:
        result = 0

    return json.dumps({"result": result, "operation": operation})


async def search_tool_impl(context: Any, arguments: str) -> str:
    """Search tool implementation."""
    args = json.loads(arguments)
    query = args.get("query", "")
    return json.dumps({"results": [f"Result 1 for '{query}'", f"Result 2 for '{query}'"]})


async def database_tool_impl(context: Any, arguments: str) -> str:
    """Database tool implementation."""
    args = json.loads(arguments)
    action = args.get("action", "query")
    return json.dumps({"status": "success", "action": action, "rows": 42})


@pytest.fixture
def multi_tool_agent():
    """Create an agent with multiple tools for testing different configurations."""
    calculator = FunctionTool(
        name="calculator",
        description="Perform mathematical operations",
        params_json_schema={
            "type": "object",
            "properties": {
                "operation": {"type": "string", "enum": ["add", "subtract", "multiply", "divide"]},
                "a": {"type": "number"},
                "b": {"type": "number"},
            },
            "required": ["operation", "a", "b"],
        },
        on_invoke_tool=calculator_impl,
    )

    search = FunctionTool(
        name="search_web",
        description="Search the web for information",
        params_json_schema={
            "type": "object",
            "properties": {
                "query": {"type": "string"},
            },
            "required": ["query"],
        },
        on_invoke_tool=search_tool_impl,
    )

    database = FunctionTool(
        name="database_query",
        description="Query a database",
        params_json_schema={
            "type": "object",
            "properties": {
                "action": {"type": "string", "enum": ["query", "update", "delete"]},
                "sql": {"type": "string"},
            },
            "required": ["action"],
        },
        on_invoke_tool=database_tool_impl,
    )

    return Agent(
        name="MultiToolAssistant",
        model="gpt-4o-mini",
        instructions="You are a helpful assistant with access to calculator, search, and database tools.",
        tools=[calculator, search, database],
    )


@pytest.fixture
def mock_api_client_with_options():
    """Mock the API client to verify ToolCallOptions are passed correctly."""
    calls = []

    def mock_post_sync(endpoint, payload):
        calls.append({"endpoint": endpoint, "payload": payload})
        return {"result": "mocked"}

    from veris_ai.api_client import get_api_client

    with patch.object(get_api_client(), "post_sync", side_effect=mock_post_sync):
        yield {"calls": calls}


@pytest.mark.asyncio
async def test_tool_options_passed_to_mock_call(mock_api_client_with_options, multi_tool_agent):
    """Verify that ToolCallOptions are correctly passed to the mock_tool_call function."""
    # Set up session for mocking using proper token format
    token = create_test_token("test-session-tool-options")
    veris.set_session_id(token)

    try:
        config = VerisConfig(
            tool_options={
                "calculator": ToolCallOptions(
                    response_expectation=ResponseExpectation.REQUIRED,
                    cache_response=True,
                    mode="function",
                )
            }
        )
        runner = Runner(veris_config=config)

        # Note: This test would require actual mocking infrastructure
        # For now, we're just verifying the configuration is created correctly
        assert (
            runner.veris_config.tool_options["calculator"].response_expectation
            == ResponseExpectation.REQUIRED
        )
        assert runner.veris_config.tool_options["calculator"].cache_response is True
        assert runner.veris_config.tool_options["calculator"].mode == "function"

    finally:
        # Clean up
        os.environ.pop("ENV", None)
        veris.clear_session_id()


def test_tool_call_options_defaults():
    """Test that ToolCallOptions has correct default values."""
    options = ToolCallOptions()
    assert options.response_expectation == ResponseExpectation.AUTO
    assert options.cache_response is False
    assert options.mode == "tool"

    # Test with custom values
    options = ToolCallOptions(
        response_expectation=ResponseExpectation.REQUIRED, cache_response=True, mode="spy"
    )
    assert options.response_expectation == ResponseExpectation.REQUIRED
    assert options.cache_response is True
    assert options.mode == "spy"


def test_response_expectation_enum():
    """Test ResponseExpectation enum values."""
    assert ResponseExpectation.AUTO == "auto"
    assert ResponseExpectation.REQUIRED == "required"
    assert ResponseExpectation.NONE == "none"

    # Verify all enum values are accessible
    all_values = [e.value for e in ResponseExpectation]
    assert "auto" in all_values
    assert "required" in all_values
    assert "none" in all_values


# ===== Tests for new Runner class with ToolCallOptions =====


@pytest.mark.asyncio
async def test_runner_class_with_tool_options_basic(multi_tool_agent):
    """Test Runner class with basic ToolCallOptions configuration."""
    veris.clear_session_id()

    # Create config with tool options
    config = VerisConfig(
        tool_options={
            "calculator": ToolCallOptions(
                response_expectation=ResponseExpectation.AUTO, cache_response=True, mode="tool"
            ),
            "search_web": ToolCallOptions(
                response_expectation=ResponseExpectation.NONE, cache_response=False, mode="spy"
            ),
        }
    )

    result = await Runner.run(
        multi_tool_agent, "Calculate 15 plus 27", veris_config=config, max_turns=2
    )

    assert result is not None
    # Check that the calculation was performed (42)
    last_message = result.to_input_list()[-1]
    if isinstance(last_message, dict) and "content" in last_message:
        content = last_message["content"]
        if isinstance(content, list) and len(content) > 0:
            text = content[0].get("text", "")
        else:
            text = str(content)
    else:
        text = str(last_message)
    assert "42" in text


@pytest.mark.asyncio
async def test_runner_with_tool_options_and_include_tools(multi_tool_agent):
    """Test Runner with both tool_options and include_tools."""
    veris.clear_session_id()

    # Create config that only intercepts calculator with specific options
    config = VerisConfig(
        include_tools=["calculator"],
        tool_options={
            "calculator": ToolCallOptions(
                response_expectation=ResponseExpectation.REQUIRED, cache_response=True
            ),
            "search_web": ToolCallOptions(  # This won't be used since not in include_tools
                response_expectation=ResponseExpectation.NONE
            ),
        },
    )

    result = await Runner.run(
        multi_tool_agent, "What is 8 times 6?", veris_config=config, max_turns=2
    )

    assert result is not None
    # Check for the result (48)
    last_message = result.to_input_list()[-1]
    if isinstance(last_message, dict) and "content" in last_message:
        content = last_message["content"]
        if isinstance(content, list) and len(content) > 0:
            text = content[0].get("text", "")
        else:
            text = str(content)
    else:
        text = str(last_message)
    assert "48" in text


def test_veris_config_with_tool_options():
    """Test that VerisConfig correctly handles tool_options."""
    # Test with empty tool_options
    config = VerisConfig()
    assert config.tool_options is None

    # Test with single tool option
    config = VerisConfig(
        tool_options={"my_tool": ToolCallOptions(response_expectation=ResponseExpectation.REQUIRED)}
    )
    assert config.tool_options is not None
    assert "my_tool" in config.tool_options
    assert config.tool_options["my_tool"].response_expectation == ResponseExpectation.REQUIRED

    # Test with multiple tool options
    config = VerisConfig(
        include_tools=["tool1", "tool2"],
        tool_options={
            "tool1": ToolCallOptions(
                response_expectation=ResponseExpectation.AUTO, cache_response=True, mode="tool"
            ),
            "tool2": ToolCallOptions(
                response_expectation=ResponseExpectation.NONE, cache_response=False, mode="spy"
            ),
        },
    )
    assert len(config.tool_options) == 2
    assert config.tool_options["tool1"].cache_response is True
    assert config.tool_options["tool2"].mode == "spy"


@pytest.mark.asyncio
async def test_runner_class_static_method_usage(multi_tool_agent):
    """Test that Runner.run can be used as a static method (OpenAI style)."""
    veris.clear_session_id()

    # Use Runner.run directly without creating an instance
    result = await Runner.run(multi_tool_agent, "Calculate 100 divided by 4", max_turns=2)

    assert result is not None
    last_message = result.to_input_list()[-1]
    if isinstance(last_message, dict) and "content" in last_message:
        content = last_message["content"]
        if isinstance(content, list) and len(content) > 0:
            text = content[0].get("text", "")
        else:
            text = str(content)
    else:
        text = str(last_message)
    assert "25" in text



"""Models for the VERIS SDK."""

from enum import Enum
from typing import Literal

from pydantic import BaseModel


class ResponseExpectation(str, Enum):
    """Expected response behavior for tool mocking."""

    AUTO = "auto"
    REQUIRED = "required"
    NONE = "none"


class ToolCallOptions(BaseModel):
    """Options for tool call."""

    response_expectation: ResponseExpectation = ResponseExpectation.AUTO
    cache_response: bool = False
    mode: Literal["tool", "function"] = "tool"



import base64
import inspect
import json
import logging
from collections.abc import Callable
from contextlib import suppress
from contextvars import ContextVar
from functools import wraps
import tenacity
from typing import (
    Any,
    Literal,
    TypeVar,
    get_type_hints,
)


from veris_ai.models import ResponseExpectation, ToolCallOptions
from veris_ai.api_client import _base_url_context, get_api_client
from veris_ai.utils import (
    convert_to_type,
    execute_callback,
    execute_combined_callback,
    extract_json_schema,
    get_function_parameters,
    get_input_parameters,
    launch_callback_task,
    launch_combined_callback_task,
)

logger = logging.getLogger(__name__)

T = TypeVar("T")

# Context variables to store session_id and thread_id for each call
_session_id_context: ContextVar[str | None] = ContextVar("veris_session_id", default=None)
_thread_id_context: ContextVar[str | None] = ContextVar("veris_thread_id", default=None)


class VerisSDK:
    """Class for mocking tool calls."""

    def __init__(self) -> None:
        """Initialize the ToolMock class."""
        self._mcp = None

    @property
    def session_id(self) -> str | None:
        """Get the session_id from context variable."""
        return _session_id_context.get()

    @property
    def thread_id(self) -> str | None:
        """Get the thread_id from context variable."""
        return _thread_id_context.get()

    def _set_session_id(self, session_id: str) -> None:
        """Set the session_id in context variable (private method)."""
        _session_id_context.set(session_id)

    def _set_thread_id(self, thread_id: str) -> None:
        """Set the thread_id in context variable (private method)."""
        _thread_id_context.set(thread_id)

    def parse_token(self, token: str) -> None:
        """Parse and set session_id and thread_id from a base64-encoded token.

        The token must be a base64-encoded JSON object containing:
        {"session_id": "...", "thread_id": "..."}

        Both session_id and thread_id are required fields.

        Args:
            token: Base64-encoded JSON token

        Raises:
            ValueError: If token is not valid or missing required fields
        """
        try:
            # Decode base64 JSON token
            decoded = base64.b64decode(token.encode("utf-8")).decode("utf-8")
            token_data = json.loads(decoded)

            if not isinstance(token_data, dict):
                raise ValueError("Token must decode to a JSON object")

            if "session_id" not in token_data:
                raise ValueError("Token must contain 'session_id' field")

            if "thread_id" not in token_data:
                raise ValueError("Token must contain 'thread_id' field")

            self._set_session_id(token_data["session_id"])
            self._set_thread_id(token_data["thread_id"])
            if token_data.get("api_url"):
                _base_url_context.set(token_data["api_url"])
            logger.info(
                f"Session ID set to {token_data['session_id']}, "
                f"Thread ID set to {token_data['thread_id']} - mocking enabled"
            )

        except (ValueError, json.JSONDecodeError) as e:
            error_msg = f"Invalid token format: {e}"
            raise ValueError(error_msg) from e

    def set_session_id(self, token: str) -> None:
        """DEPRECATED: Use parse_token() instead.

        Set the session_id and thread_id from a base64-encoded token.

        Args:
            token: Base64-encoded JSON token
        """
        logger.warning(
            "set_session_id() is deprecated. Use parse_token() instead. "
            "This method will be removed in a future version."
        )
        # For backwards compatibility, allow tokens without thread_id
        try:
            decoded = base64.b64decode(token.encode("utf-8")).decode("utf-8")
            token_data = json.loads(decoded)

            if not isinstance(token_data, dict):
                raise ValueError("Token must decode to a JSON object")

            if "session_id" not in token_data:
                raise ValueError("Token must contain 'session_id' field")

            self._set_session_id(token_data["session_id"])
            if "thread_id" in token_data:
                self._set_thread_id(token_data["thread_id"])
            logger.info(f"Session ID set to {token_data['session_id']}")

        except (ValueError, json.JSONDecodeError) as e:
            error_msg = f"Invalid token format: {e}"
            raise ValueError(error_msg) from e

    def _clear_session_id(self) -> None:
        """Clear the session_id from context variable (private method)."""
        _session_id_context.set(None)

    def _clear_thread_id(self) -> None:
        """Clear the thread_id from context variable (private method)."""
        _thread_id_context.set(None)

    def clear_context(self) -> None:
        """Clear the session_id, thread_id, and base_url from context variables."""
        self._clear_session_id()
        self._clear_thread_id()
        _base_url_context.set(None)
        logger.info("Session ID, Thread ID, and Base URL cleared - mocking disabled")

    def clear_session_id(self) -> None:
        """DEPRECATED: Use clear_context() instead."""
        logger.warning(
            "clear_session_id() is deprecated. Use clear_context() instead. "
            "This method will be removed in a future version."
        )
        self.clear_context()

    @property
    def fastapi_mcp(self) -> Any | None:  # noqa: ANN401
        """Get the FastAPI MCP server."""
        return self._mcp

    def set_fastapi_mcp(self, **params_dict: Any) -> None:  # noqa: ANN401
        """Set the FastAPI MCP server with HTTP transport."""
        from fastapi import Depends, Request  # noqa: PLC0415
        from fastapi.security import OAuth2PasswordBearer  # noqa: PLC0415
        from fastapi_mcp import (  # type: ignore[import-untyped] # noqa: PLC0415
            AuthConfig,
            FastApiMCP,
        )

        oauth2_scheme = OAuth2PasswordBearer(tokenUrl="token", auto_error=False)

        async def authenticate_request(
            request: Request,  # noqa: ARG001
            token: str | None = Depends(oauth2_scheme),
        ) -> None:
            if token:
                self.parse_token(token)

        # Create auth config with dependencies
        auth_config = AuthConfig(
            dependencies=[Depends(authenticate_request)],
        )

        # Merge the provided params with our auth config
        if "auth_config" in params_dict:
            # Merge the provided auth config with our dependencies
            provided_auth_config = params_dict.pop("auth_config")
            if provided_auth_config.dependencies:
                auth_config.dependencies.extend(provided_auth_config.dependencies)
            # Copy other auth config properties if they exist
            for field, value in provided_auth_config.model_dump(exclude_none=True).items():
                if field != "dependencies" and hasattr(auth_config, field):
                    setattr(auth_config, field, value)

        # Create the FastApiMCP instance with merged parameters
        self._mcp = FastApiMCP(
            auth_config=auth_config,
            **params_dict,
        )

    def spy(self) -> Callable:
        """Decorator for spying on tool calls."""

        def decorator(func: Callable) -> Callable:
            """Decorator for spying on tool calls."""
            is_async = inspect.iscoroutinefunction(func)

            @wraps(func)
            async def async_wrapper(*args: tuple[object, ...], **kwargs: Any) -> object:  # noqa: ANN401
                """Async wrapper."""
                session_id = _session_id_context.get()
                if not session_id:
                    return await func(*args, **kwargs)
                parameters = get_function_parameters(func, args, kwargs)
                logger.info(f"Spying on function: {func.__name__}")
                await log_tool_call_async(
                    session_id=session_id,
                    function_name=func.__name__,
                    parameters=parameters,
                    docstring=inspect.getdoc(func) or "",
                )
                result = await func(*args, **kwargs)
                await log_tool_response_async(session_id=session_id, response=result)
                return result

            @wraps(func)
            def sync_wrapper(*args: tuple[object, ...], **kwargs: Any) -> object:  # noqa: ANN401
                """Sync wrapper."""
                session_id = _session_id_context.get()
                if not session_id:
                    return func(*args, **kwargs)
                parameters = get_function_parameters(func, args, kwargs)
                logger.info(f"Spying on function: {func.__name__}")
                log_tool_call(
                    session_id=session_id,
                    function_name=func.__name__,
                    parameters=parameters,
                    docstring=inspect.getdoc(func) or "",
                )
                result = func(*args, **kwargs)
                log_tool_response(session_id=session_id, response=result)
                return result

            return async_wrapper if is_async else sync_wrapper

        return decorator

    def mock(  # noqa: C901, PLR0915, PLR0913
        self,
        mode: Literal["tool", "function"] = "tool",
        expects_response: bool | None = None,
        cache_response: bool | None = None,
        input_callback: Callable[..., Any] | None = None,
        output_callback: Callable[[Any], Any] | None = None,
        combined_callback: Callable[..., Any] | None = None,
    ) -> Callable:
        """Decorator for mocking tool calls.

        Args:
            mode: Whether to treat the function as a tool or function
            expects_response: Whether the function expects a response
            cache_response: Whether to cache the response
            input_callback: Callable that receives input parameters as individual arguments
            output_callback: Callable that receives the output value
            combined_callback: Callable that receives both input parameters and mock_output
        """
        response_expectation = (
            ResponseExpectation.NONE
            if (expects_response is False or (expects_response is None and mode == "function"))
            else ResponseExpectation.REQUIRED
            if expects_response is True
            else ResponseExpectation.AUTO
        )
        cache_response = cache_response or False
        options = ToolCallOptions(
            mode=mode, response_expectation=response_expectation, cache_response=cache_response
        )

        def decorator(func: Callable) -> Callable:  # noqa: C901, PLR0915
            """Decorator for mocking tool calls."""
            is_async = inspect.iscoroutinefunction(func)

            @wraps(func)
            async def async_wrapper(
                *args: tuple[object, ...],
                **kwargs: Any,  # noqa: ANN401
            ) -> object:
                """Async wrapper."""
                session_id = _session_id_context.get()
                if not session_id:
                    logger.info(
                        f"No session ID found, executing original function: {func.__name__}"
                    )
                    return await func(*args, **kwargs)

                # Perform the mock call first
                parameters = get_function_parameters(func, args, kwargs)
                thread_id = _thread_id_context.get()
                result = await mock_tool_call_async(
                    func,
                    session_id,
                    parameters,
                    options,
                    thread_id,
                )

                # Launch callbacks as background tasks (non-blocking)
                input_params = get_input_parameters(func, args, kwargs)
                launch_callback_task(input_callback, input_params, unpack=True)
                launch_callback_task(output_callback, result, unpack=False)
                launch_combined_callback_task(combined_callback, input_params, result)

                return result

            @wraps(func)
            def sync_wrapper(
                *args: tuple[object, ...],
                **kwargs: Any,  # noqa: ANN401
            ) -> object:
                """Sync wrapper."""
                session_id = _session_id_context.get()
                if not session_id:
                    logger.info(
                        f"No session ID found, executing original function: {func.__name__}"
                    )
                    return func(*args, **kwargs)

                # Perform the mock call first
                parameters = get_function_parameters(func, args, kwargs)
                thread_id = _thread_id_context.get()
                result = mock_tool_call(
                    func,
                    session_id,
                    parameters,
                    options,
                    thread_id,
                )

                # Execute callbacks synchronously (can't use async tasks in sync context)
                input_params = get_input_parameters(func, args, kwargs)
                execute_callback(input_callback, input_params, unpack=True)
                execute_callback(output_callback, result, unpack=False)
                execute_combined_callback(combined_callback, input_params, result)

                return result

            # Return the appropriate wrapper based on whether the function is async
            return async_wrapper if is_async else sync_wrapper

        return decorator

    def stub(
        self,
        return_value: Any,  # noqa: ANN401
        input_callback: Callable[..., Any] | None = None,
        output_callback: Callable[[Any], Any] | None = None,
        combined_callback: Callable[..., Any] | None = None,
    ) -> Callable:
        """Decorator for stubbing tool calls.

        Args:
            return_value: The value to return when the function is stubbed
            input_callback: Callable that receives input parameters as individual arguments
            output_callback: Callable that receives the output value
            combined_callback: Callable that receives both input parameters and mock_output
        """

        def decorator(func: Callable) -> Callable:
            # Check if the original function is async
            is_async = inspect.iscoroutinefunction(func)

            @wraps(func)
            async def async_wrapper(
                *args: tuple[object, ...],
                **kwargs: Any,  # noqa: ANN401
            ) -> object:
                if not self.session_id:
                    logger.info(
                        f"No session ID found, executing original function: {func.__name__}"
                    )
                    return await func(*args, **kwargs)

                logger.info(f"Stubbing function: {func.__name__}")

                # Launch callbacks as background tasks (non-blocking)
                input_params = get_input_parameters(func, args, kwargs)
                launch_callback_task(input_callback, input_params, unpack=True)
                launch_callback_task(output_callback, return_value, unpack=False)
                launch_combined_callback_task(combined_callback, input_params, return_value)

                return return_value

            @wraps(func)
            def sync_wrapper(*args: tuple[object, ...], **kwargs: Any) -> object:  # noqa: ANN401
                if not self.session_id:
                    logger.info(
                        f"No session ID found, executing original function: {func.__name__}"
                    )
                    return func(*args, **kwargs)

                logger.info(f"Stubbing function: {func.__name__}")

                # Execute callbacks synchronously (can't use async tasks in sync context)
                input_params = get_input_parameters(func, args, kwargs)
                execute_callback(input_callback, input_params, unpack=True)
                execute_callback(output_callback, return_value, unpack=False)
                execute_combined_callback(combined_callback, input_params, return_value)

                return return_value

            # Return the appropriate wrapper based on whether the function is async
            return async_wrapper if is_async else sync_wrapper

        return decorator


@tenacity.retry(
    stop=tenacity.stop_after_attempt(3),
    wait=tenacity.wait_exponential(multiplier=1, min=4, max=10),
    reraise=True,
)
def mock_tool_call(
    func: Callable,
    session_id: str,  # noqa: ARG001
    parameters: dict[str, dict[str, str]],
    options: ToolCallOptions | None = None,
    thread_id: str | None = None,
) -> object:
    """Mock tool call (synchronous).

    Args:
        func: Function being mocked
        session_id: Session ID (kept for backwards compatibility, not used)
        parameters: Function parameters
        options: Tool call options
        thread_id: Thread ID to use as session_id in API request (required)

    Raises:
        ValueError: If thread_id is not provided
    """
    if thread_id is None:
        raise ValueError(
            "thread_id is required for mocking. "
            "Use parse_token() to set both session_id and thread_id."
        )

    options = options or ToolCallOptions()
    api_client = get_api_client()
    endpoint = api_client.tool_mock_endpoint

    logger.info(f"Simulating function: {func.__name__}")

    type_hints = get_type_hints(func)

    # Extract return type object (not just the name)
    return_type_obj = type_hints.pop("return", Any)
    # Get function docstring
    docstring = inspect.getdoc(func) or ""

    # Use thread_id as session_id in the payload
    payload_session_id = thread_id
    # Clean up parameters for V3 - just send values, not the nested dict
    clean_params: dict[str, Any] = {}
    for key, value in parameters.items():
        if isinstance(value, dict) and "value" in value:
            # Extract just the value from the nested structure
            clean_params[key] = value["value"]
        else:
            # Already clean or unexpected format
            clean_params[key] = value

    # Determine response expectation
    payload = {
        "session_id": payload_session_id,
        "response_expectation": options.response_expectation.value,
        "cache_response": bool(options.cache_response),
        "tool_call": {
            "function_name": func.__name__,
            "parameters": clean_params,
            "return_type": json.dumps(extract_json_schema(return_type_obj)),
            "docstring": docstring,
        },
    }

    mock_result = api_client.post(endpoint, payload)
    logger.info(f"Mock response: {mock_result}")

    if isinstance(mock_result, str):
        with suppress(json.JSONDecodeError):
            mock_result = json.loads(mock_result)
            return convert_to_type(mock_result, return_type_obj)
    return convert_to_type(mock_result, return_type_obj)


@tenacity.retry(
    stop=tenacity.stop_after_attempt(3),
    wait=tenacity.wait_exponential(multiplier=1, min=4, max=10),
    reraise=True,
)
async def mock_tool_call_async(
    func: Callable,
    session_id: str,  # noqa: ARG001
    parameters: dict[str, dict[str, str]],
    options: ToolCallOptions | None = None,
    thread_id: str | None = None,
) -> object:
    """Mock tool call (asynchronous).

    Args:
        func: Function being mocked
        session_id: Session ID (kept for backwards compatibility, not used)
        parameters: Function parameters
        options: Tool call options
        thread_id: Thread ID to use as session_id in API request (required)

    Raises:
        ValueError: If thread_id is not provided
    """
    if thread_id is None:
        raise ValueError(
            "thread_id is required for mocking. "
            "Use parse_token() to set both session_id and thread_id."
        )

    options = options or ToolCallOptions()
    api_client = get_api_client()
    endpoint = api_client.tool_mock_endpoint

    logger.info(f"Simulating function: {func.__name__}")

    type_hints = get_type_hints(func)

    # Extract return type object (not just the name)
    return_type_obj = type_hints.pop("return", Any)
    # Get function docstring
    docstring = inspect.getdoc(func) or ""

    # Use thread_id as session_id in the payload
    payload_session_id = thread_id
    # Clean up parameters for V3 - just send values, not the nested dict
    clean_params: dict[str, Any] = {}
    for key, value in parameters.items():
        if isinstance(value, dict) and "value" in value:
            # Extract just the value from the nested structure
            clean_params[key] = value["value"]
        else:
            # Already clean or unexpected format
            clean_params[key] = value

    # Determine response expectation
    payload = {
        "session_id": payload_session_id,
        "response_expectation": options.response_expectation.value,
        "cache_response": bool(options.cache_response),
        "tool_call": {
            "function_name": func.__name__,
            "parameters": clean_params,
            "return_type": json.dumps(extract_json_schema(return_type_obj)),
            "docstring": docstring,
        },
    }

    mock_result = await api_client.post_async(endpoint, payload)
    logger.info(f"Mock response: {mock_result}")

    if isinstance(mock_result, str):
        with suppress(json.JSONDecodeError):
            mock_result = json.loads(mock_result)
            return convert_to_type(mock_result, return_type_obj)
    return convert_to_type(mock_result, return_type_obj)


def log_tool_call(
    session_id: str,
    function_name: str,
    parameters: dict[str, dict[str, str]],
    docstring: str,
) -> None:
    """Log tool call synchronously to the VERIS logging endpoint."""
    api_client = get_api_client()
    endpoint = api_client.get_log_tool_call_endpoint(session_id)

    # Clean up parameters for V3 - just send values, not the nested dict
    clean_params: dict[str, Any] = {}
    for key, value in parameters.items():
        if isinstance(value, dict) and "value" in value:
            # Extract just the value from the nested structure
            clean_params[key] = value["value"]
        else:
            # Already clean or unexpected format
            clean_params[key] = value

    payload = {
        "function_name": function_name,
        "parameters": clean_params,
        "docstring": docstring,
    }
    try:
        api_client.post(endpoint, payload)
        logger.debug(f"Tool call logged for {function_name}")
    except Exception as e:
        logger.warning(f"Failed to log tool call for {function_name}: {e}")


async def log_tool_call_async(
    session_id: str,
    function_name: str,
    parameters: dict[str, dict[str, str]],
    docstring: str,
) -> None:
    """Log tool call asynchronously to the VERIS logging endpoint."""
    api_client = get_api_client()
    endpoint = api_client.get_log_tool_call_endpoint(session_id)

    # Clean up parameters for V3 - just send values, not the nested dict
    clean_params: dict[str, Any] = {}
    for key, value in parameters.items():
        if isinstance(value, dict) and "value" in value:
            # Extract just the value from the nested structure
            clean_params[key] = value["value"]
        else:
            # Already clean or unexpected format
            clean_params[key] = value

    payload = {
        "function_name": function_name,
        "parameters": clean_params,
        "docstring": docstring,
    }
    try:
        await api_client.post_async(endpoint, payload)
        logger.debug(f"Tool call logged for {function_name}")
    except Exception as e:
        logger.warning(f"Failed to log tool call for {function_name}: {e}")


def log_tool_response(session_id: str, response: Any) -> None:  # noqa: ANN401
    """Log tool response synchronously to the VERIS logging endpoint."""
    api_client = get_api_client()
    endpoint = api_client.get_log_tool_response_endpoint(session_id)

    payload = {
        "response": json.dumps(response, default=str),
    }

    try:
        api_client.post(endpoint, payload)
        logger.debug("Tool response logged")
    except Exception as e:
        logger.warning(f"Failed to log tool response: {e}")


async def log_tool_response_async(session_id: str, response: Any) -> None:  # noqa: ANN401
    """Log tool response asynchronously to the VERIS logging endpoint."""
    api_client = get_api_client()
    endpoint = api_client.get_log_tool_response_endpoint(session_id)

    payload = {
        "response": json.dumps(response, default=str),
    }

    try:
        await api_client.post_async(endpoint, payload)
        logger.debug("Tool response logged")
    except Exception as e:
        logger.warning(f"Failed to log tool response: {e}")


veris = VerisSDK()
